"""
Kompos helper utilities.
"""
from kompos.helpers.console import (
    Colors,
    print_error,
    print_success,
    print_info,
    print_warning,
    print_section_header,
    print_subsection,
    print_separator,
    print_summary,
    print_kvp,
    print_composition_header,
    print_file_generation,
)

__all__ = [
    'Colors',
    'print_error',
    'print_success',
    'print_info',
    'print_warning',
    'print_section_header',
    'print_subsection',
    'print_separator',
    'print_summary',
    'print_kvp',
    'print_composition_header',
    'print_file_generation',
]



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import sys

from kompos import __version__
from kompos.helpers import print_error


class KomposArgumentParser(argparse.ArgumentParser):
    """Custom ArgumentParser with better error messages."""
    
    def error(self, message):
        """Override error to provide cleaner, more helpful messages."""
        # Extract just the essential error message
        if 'required: subcommand' in message.lower() or 'invalid choice' in message.lower():
            runner_name = self.prog.split()[-1] if len(self.prog.split()) > 1 else 'unknown'
            
            # Try to extract choices from the parser (check both subparsers and regular arguments with choices)
            choices = []
            for action in self._actions:
                if isinstance(action, argparse._SubParsersAction):
                    choices = list(action.choices.keys()) if action.choices else []
                    break
                elif hasattr(action, 'choices') and action.choices and action.dest in ('subcommand', 'command'):
                    choices = list(action.choices)
                    break
            
            if choices:
                print_error(f"Missing or invalid subcommand for '{runner_name}'")
                print("", file=sys.stderr)
                print(f"Available commands: {', '.join(choices)}", file=sys.stderr)
                print("", file=sys.stderr)
                print(f"Usage: kompos <config_path> {runner_name} {{{','.join(choices)}}}", file=sys.stderr)
            else:
                print_error(f"Missing required subcommand for '{runner_name}'")
                print("", file=sys.stderr)
                print(f"Run 'kompos <config_path> {runner_name} --help' to see available commands.", file=sys.stderr)
        else:
            print_error(message)
            print("", file=sys.stderr)
            self.print_usage(sys.stderr)
        sys.exit(2)


class RootParser:
    def __init__(self, sub_parsers=None):
        """
        :type sub_parsers: list[SubParserConfig]
        """

        if sub_parsers is None:
            sub_parsers = []
        self.sub_parsers = sub_parsers

    def _get_parser(self):
        parser = KomposArgumentParser(
            description='Run commands against a definition', prog='kompos')
        parser.add_argument('config_path',
                            type=str,
                            help='The config path from where to run compositions or generate a flat config file.')
        parser.add_argument('--root-path',
                            type=str,
                            help='The root of the resource tree it can be an absolute path or relative to the current dir')
        parser.add_argument('--verbose', '-v', action='count',
                            help='Get more verbose output from commands')
        parser.add_argument('--version',
                            action='version',
                            version=f'%(prog)s v{__version__}'
                            )
        parser.add_argument('--himl',
                            action='store',
                            dest='himl_args',
                            default=None,
                            help='for passing arguments to himl'
                                 '--himl="--arg1 --arg2" any himl argument is supported wrapped in quotes')
        subparsers = parser.add_subparsers(dest='command', parser_class=KomposArgumentParser)

        for subparser_conf in self.sub_parsers:
            subparser_instance = subparsers.add_parser(subparser_conf.get_name(),
                                                       help=subparser_conf.get_help(),
                                                       epilog=subparser_conf.get_epilog(),
                                                       formatter_class=subparser_conf.get_formatter())
            subparser_conf.configure(subparser_instance)

        return parser

    @staticmethod
    def _check_args_for_unicode(args):
        if args is None:
            args = sys.argv
        try:
            for value in args:
                value.encode('utf-8')
        except UnicodeDecodeError as e:
            print(
                f'Invalid character in argument "{e.args[1]}", most likely an "en dash", replace it with normal dash -')
            raise

    def parse_args(self, args=None):
        RootParser._check_args_for_unicode(args)
        return self._get_parser().parse_args(args)

    def parse_known_args(self, args=None):
        RootParser._check_args_for_unicode(args)
        return self._get_parser().parse_known_args(args)


class SubParserConfig:
    def get_name(self):
        pass

    def configure(self, parser):
        """
        Configure parser arguments for this runner.
        
        Override this method to add runner-specific arguments.
        Call self.add_himl_arguments(parser) to add standard HIML arguments.
        
        Args:
            parser: The argument parser to configure
        
        Returns:
            The configured parser
        """
        pass

    def get_formatter(self):
        return argparse.RawDescriptionHelpFormatter

    def get_help(self):
        return ""

    def get_epilog(self):
        return ""
    
    def add_himl_arguments(self, parser):
        """
        Helper to add all HIML arguments to the parser.
        
        Call this from your configure() method if your runner needs HIML args.
        Most runners that call generate_config() will need this.
        
        Args:
            parser: The argument parser to add HIML arguments to
        
        Returns:
            The parser with HIML arguments added
        
        Example:
            def configure(self, parser):
                parser.add_argument('--my-flag', help='My custom flag')
                self.add_himl_arguments(parser)  # Add standard HIML args
                return parser
        """
        from himl import ConfigRunner
        return ConfigRunner().get_parser(parser)



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import fcntl
import logging
import os
import shutil
from subprocess import Popen, PIPE

from himl import ConfigRunner
from himl.config_generator import ConfigProcessor
from himl.interpolation import InterpolationResolver

logger = logging.getLogger(__name__)

COMPOSITION_KEY = "composition"


class GenericRunner:
    """
    Base class for all Kompos runners.
    
    Provides:
    - HIML config generation via ConfigProcessor
    - Composition discovery and orchestration
    - Common utilities for all runner types (terraform, helmfile, tfe, config, explore)
    """

    def __init__(self, kompos_config, config_path, execute, runner_type):
        # Initialize HIML config processor
        self.config_processor = ConfigProcessor()

        logging.basicConfig(level=logging.INFO)

        self.execute = execute
        self.runner_type = runner_type
        self.validate_runner = True

        self.kompos_config = kompos_config
        self.config_path = config_path
        self.himl_args = None
        self.reverse = False
        self.ordered_compositions = False

        self.generate_output = True

    @staticmethod
    def extract_format_from_extension(extension):
        """
        Extract output format (yaml/json) from file extension.
        
        Args:
            extension: File extension (e.g., '.workspace.yaml', '.tfvars.json')
        
        Returns:
            'yaml' or 'json'
        
        Examples:
            '.workspace.yaml' -> 'yaml'
            '.tfvars.yaml' -> 'yaml'
            '.json' -> 'json'
            '.workspace.json' -> 'json'
        """
        if 'yaml' in extension.lower() or 'yml' in extension.lower():
            return 'yaml'
        elif 'json' in extension.lower():
            return 'json'
        # Default to yaml if uncertain
        return 'yaml'

    def generate_config(
            self,
            config_path,
            filters=(),
            exclude_keys=(),
            enclosing_key=None,
            remove_enclosing_key=None,
            output_format="yaml",
            print_data=False,
            output_file=None,
            skip_interpolation_resolving=False,
            skip_interpolation_validation=False,
            skip_secrets=False,
            multi_line_string=False,
            type_strategies=[(list, ["append"]), (dict, ["merge"])],
            fallback_strategies=["override"],
            type_conflict_strategies=["override"],
            silent=False  # Don't print command (for internal/trace operations)
    ):
        """Generate hierarchical configuration using HIML."""
        cmd = self.get_sh_command(
            config_path,
            filters,
            exclude_keys,
            enclosing_key,
            remove_enclosing_key,
            output_format,
            print_data,
            output_file,
            skip_interpolation_resolving,
            skip_interpolation_validation,
            skip_secrets,
            multi_line_string,
        )

        if not silent:
            logger.debug(cmd)

        return self.config_processor.process(
            path=config_path,
            filters=filters,
            exclude_keys=exclude_keys,
            enclosing_key=enclosing_key,
            remove_enclosing_key=remove_enclosing_key,
            output_format=output_format,
            output_file=output_file,
            print_data=print_data,
            skip_interpolations=skip_interpolation_resolving,
            skip_interpolation_validation=skip_interpolation_validation,
            skip_secrets=skip_secrets,
            multi_line_string=multi_line_string,
            type_strategies=type_strategies,
            fallback_strategies=fallback_strategies,
            type_conflict_strategies=type_conflict_strategies
        )

    @staticmethod
    def get_sh_command(
            config_path,
            filters=(),
            exclude_keys=(),
            enclosing_key=None,
            remove_enclosing_key=None,
            output_format="yaml",
            print_data=False,
            output_file=None,
            skip_interpolation_resolving=False,
            skip_interpolation_validation=False,
            skip_secrets=False,
            multi_line_string=False,
    ):
        """Build shell command string for displaying HIML invocation."""
        command = f"kompos {config_path} config --format {output_format}"
        for filter in filters:
            command += f" --filter {filter}"
        for exclude in exclude_keys:
            command += f" --exclude {exclude}"
        if enclosing_key:
            command += f" --enclosing-key {enclosing_key}"
        if remove_enclosing_key:
            command += f" --remove-enclosing-key {remove_enclosing_key}"
        if output_file:
            command += f" --output-file {output_file}"
        if print_data:
            command += " --print-data"
        if skip_interpolation_resolving:
            command += " --skip-interpolation-resolving"
        if skip_interpolation_validation:
            command += " --skip-interpolation-validation"
        if skip_secrets:
            command += " --skip-secrets"
        if multi_line_string:
            command += " --multi-line-string"

        return command

    def run(self, args, extra_args):
        logger.debug(f"Runner: {self.runner_type}")

        if len(extra_args) > 1:
            logger.debug(f"Found extra_args {extra_args}")

        self.himl_args = get_himl_args(args)
        self.run_configuration(args)

        # Stop processing if an incompatible runner version is detected.
        if not self.runner_type:
            logger.error("Could not detect runner type and version.")
            exit(1)
        if self.validate_runner:
            validate_runner_version(self.kompos_config, self.runner_type)

        compositions, paths = self.get_compositions()

        return self.run_compositions(args, extra_args, compositions, paths)

    def run_configuration(self, args):
        return

    def get_compositions(self):
        logging.basicConfig(level=logging.INFO)

        compositions, paths = discover_compositions(
            self.config_path,
            self.kompos_config,
            self.runner_type
        )
        if self.ordered_compositions:
            composition_order = self.kompos_config.composition_order(self.runner_type)
            compositions = sorted_compositions(compositions, composition_order, self.reverse)

        if not compositions:
            raise Exception(
                f"No {self.runner_type} compositions were detected in {self.config_path}.")

        return compositions, paths

    def get_raw_config(self, config_path, composition):
        """
        Generate raw config WITHOUT exclusions/filters for internal use.
        
        This config is used by Kompos to read metadata like composition.instance.
        All keys must be available for interpolation (exclusions break references).
        
        Exclusions/filters are applied separately when generating final output files.
        """
        return self.generate_config(
            config_path=config_path,
            exclude_keys=[],  # No exclusions - all keys available for interpolation
            filters=[],  # No filters - generate complete config
            skip_interpolation_validation=True,
            skip_secrets=True
        )

    def get_composition_name(self, raw_config):
        """
        Get the resolved composition instance identifier from layered config.
        
        Reads composition.instance which uses pure Himl interpolation:
          composition.instance: "{{account.name}}"  or  "{{cluster.fullName}}"
        
        Args:
            raw_config: Raw configuration dictionary from himl
        
        Returns:
            str: Resolved instance identifier (e.g., "my-account", "my-cluster-us-east-1")
            
        Example:
            name = self.get_composition_name(raw_config)
            if not name:
                return  # Skip processing
        """
        return self.kompos_config.get_composition_name(
            raw_config=raw_config,
            get_nested_value_fn=self.get_nested_value
        )

    def get_composition_output_dir(self, raw_config):
        """
        Get the composition instance directory name from layered config.
        
        Reads composition.output_dir (or falls back to composition.instance) which uses
        pure Himl interpolation:
          composition.instance: "{{account.name}}"  or  "{{cluster.fullName}}"
        
        This is Level 3 in the output hierarchy:
          base/composition_type/composition_instance/
        
        Args:
            raw_config: Raw configuration dictionary from himl
        
        Returns:
            str: Resolved directory name (e.g., "my-account", "my-cluster-us-east-1")
        """
        # Try composition.output_dir first (for cases where name and dir differ)
        output_dir = self.get_nested_value(raw_config, 'composition.output_dir')
        if not output_dir or '{{' in str(output_dir):
            # Fallback to composition.instance (typical case - instance = dir)
            output_dir = self.get_composition_name(raw_config)
        return output_dir

    def run_compositions(self, args, extra_args, compositions, paths):
        # Acquire lock to prevent concurrent kompos runs from interfering
        lock_file = None
        lock_fd = None

        try:
            lock_file = os.path.join(os.getcwd(), '.kompos-runtime', '.lock')
            os.makedirs(os.path.dirname(lock_file), exist_ok=True)
            lock_fd = open(lock_file, 'w')

            # Try to acquire exclusive lock (non-blocking)
            try:
                fcntl.flock(lock_fd.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                logger.debug(f'Acquired kompos lock: {lock_file}')
            except IOError:
                logger.error('Another kompos process is already running. Please wait or kill the other process.')
                logger.error(f'Lock file: {lock_file}')
                return 1

            return self._run_compositions_internal(args, extra_args, compositions, paths)

        finally:
            # Release lock and cleanup
            if lock_fd:
                try:
                    fcntl.flock(lock_fd.fileno(), fcntl.LOCK_UN)
                    lock_fd.close()
                    logger.debug('Released kompos lock')
                except Exception as e:
                    logger.warning(f'Failed to release lock: {e}')

    def _run_compositions_internal(self, args, extra_args, compositions, paths):
        for composition in compositions:
            logger.debug(f"→ {self.runner_type.upper()}: {composition}")

            # Set current path
            config_path = paths[composition]

            # Raw config generation
            raw_config = self.get_raw_config(config_path, composition)

            # Generate output paths for configs
            default_output_path = None
            if self.generate_output:
                default_output_path = get_default_output_path(args, raw_config, self.kompos_config, self.runner_type)

            # Set default key filters
            filtered_keys = self.kompos_config.filtered_output_keys(composition)
            excluded_keys = self.kompos_config.excluded_config_keys(composition)

            # Add CLI-provided filters and excludes to config-based ones
            # Use getattr: subcommands without HIML args (e.g. helm list) won't have these attrs
            if getattr(self.himl_args, 'filter', None):
                filtered_keys = filtered_keys + self.himl_args.filter
            if getattr(self.himl_args, 'exclude', None):
                excluded_keys = excluded_keys + self.himl_args.exclude

            # Runner pre-configuration
            self.execution_configuration(composition, config_path, default_output_path, raw_config,
                                         filtered_keys, excluded_keys)

            # Execute runner
            return_code = self.execute(self.execution(args, extra_args, default_output_path, composition, raw_config))
            if return_code != 0:
                logger.error(
                    f"Command finished with nonzero exit code for composition '{composition}'."
                    f"Will skip remaining compositions."
                )
                return return_code

            # Run some code after execution
            self.execution_post_action()

        return 0

    def execution_configuration(self, composition, config_path, default_output_path, raw_config,
                                filtered_keys, excluded_keys):
        return

    @staticmethod
    def execution(args, extra_args, default_output_path, composition, raw_config):
        return

    @staticmethod
    def execution_post_action():
        return

    @staticmethod
    def get_nested_value(data, key_path, default=None):
        """
        Get a nested value from a dict using a dotted key path.
        Supports both dictionary keys and list indices.
        
        Args:
            data: Dictionary to extract value from
            key_path: Dot-separated path to the value (e.g., 'cluster.fullName', 'workspaces.0.name')
            default: Default value to return if not found (default: None)
        
        Returns:
            The value at the key path, or default if not found
        
        Examples:
            >>> data = {'cluster': {'fullName': 'my-cluster', 'name': 'cluster1'}}
            >>> GenericRunner.get_nested_value(data, 'cluster.fullName')
            'my-cluster'
            >>> data = {'workspaces': [{'name': 'ws1'}, {'name': 'ws2'}]}
            >>> GenericRunner.get_nested_value(data, 'workspaces.0.name')
            'ws1'
            >>> GenericRunner.get_nested_value(data, 'cluster.missing')
            None
            >>> GenericRunner.get_nested_value(data, 'cluster.missing', 'default-cluster')
            'default-cluster'
        """
        if not key_path:
            return default

        keys = key_path.split('.')
        value = data

        for key in keys:
            # Try to use as array index if it's a digit
            if isinstance(value, list) and key.isdigit():
                idx = int(key)
                if 0 <= idx < len(value):
                    value = value[idx]
                else:
                    return default
            # Otherwise use as dict key
            elif isinstance(value, dict) and key in value:
                value = value[key]
            else:
                return default

        return value

    def ensure_directory(self, path, is_file_path=False, clean=False):
        """
        Ensure a directory exists, with optional cleaning before creation.

        Args:
            path:         Directory path, or file path if is_file_path=True
            is_file_path: If True, extracts the parent directory from path
            clean:        If True, removes existing directory first (for fresh state)
        Returns:
            The directory path that was ensured
        """
        dir_path = os.path.dirname(path) if is_file_path else path
        if clean and os.path.exists(dir_path):
            shutil.rmtree(dir_path)
        os.makedirs(dir_path, exist_ok=True)
        return dir_path

    @staticmethod
    def resolve_interpolations(data):
        """
        Resolve {{key.path}} interpolations in an in-memory dict.

        Runs two passes to handle nested interpolations ({{outer.{{inner}}}}).
        Used when working with pre-loaded dicts rather than filesystem paths
        (use generate_config() for filesystem-based resolution).

        Args:
            data: Dict to resolve in-place. Modified directly.
        """
        InterpolationResolver().resolve_interpolations(data)
        InterpolationResolver().resolve_interpolations(data)

    @staticmethod
    def flatten_dict(d, parent_key='', sep='.'):
        """
        Flatten nested dictionary into dot-notation keys.
        
        Args:
            d: Dictionary to flatten
            parent_key: Parent key for recursion
            sep: Separator for key concatenation (default: '.')
        
        Returns:
            Flattened dictionary with dot-notation keys
        
        Examples:
            >>> data = {'vpc': {'cidr': '10.0.0.0/16'}}
            >>> GenericRunner.flatten_dict(data)
            {'vpc.cidr': '10.0.0.0/16'}
            >>> data = {'cluster': {'name': 'prod', 'region': 'us-east-1'}}
            >>> GenericRunner.flatten_dict(data)
            {'cluster.name': 'prod', 'cluster.region': 'us-east-1'}
        """
        items = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(GenericRunner.flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)


def discover_compositions(config_path, kompos_config=None, runner_type=None):
    path_params = dict(split_path(x) for x in config_path.split('/'))

    # Check if composition is directly specified in path (e.g., composition=cluster)
    composition = path_params.get(COMPOSITION_KEY, None)
    if composition:
        # Single composition specified - return it directly
        return [composition], {composition: config_path}

    # No composition specified - discover from filesystem or config
    # This happens when path ends at cluster level (e.g., cluster=demo-cluster-01)
    logger.info("No composition specified in path. Discovering compositions...")

    # Try to discover composition paths from filesystem
    paths = {}
    compositions = []
    for subpath in os.listdir(config_path):
        if COMPOSITION_KEY + "=" in subpath:
            comp_name = split_path(subpath)[1]
            comp_path = os.path.join(config_path, subpath)
            paths[comp_name] = comp_path
            compositions.append(comp_name)

    # If nothing found on filesystem, fallback to .komposconfig.yaml
    if not compositions and kompos_config and runner_type:
        config_compositions = kompos_config.composition_order(runner_type, default=None)
        if config_compositions:
            logger.info(f"No compositions found on filesystem. Using .komposconfig.yaml: {config_compositions}")
            for comp in config_compositions:
                comp_path = os.path.join(config_path, f"{COMPOSITION_KEY}={comp}")
                paths[comp] = comp_path
                if not os.path.exists(comp_path):
                    logger.warning(f"Composition {comp} defined in config but path not found: {comp_path}")
            return config_compositions, paths

    if not compositions:
        # No composition in path and none discovered - return dummy for config rendering
        logger.warning("No composition detected in path. Config will be rendered at this level.")
        return ['config'], {'config': config_path}

    return compositions, paths


def sorted_compositions(compositions, composition_order, reverse=False):
    result = list(filter(lambda x: x in compositions, composition_order))
    return tuple(reversed(result)) if reverse else result


def split_path(value, separator='='):
    if separator in value:
        return value.split(separator)
    return [value, ""]


def get_default_output_path(args, raw_config, kompos_config, runner):
    # Use a dedicated runtime directory (not versioned)
    # This keeps generated files separate from source files
    runtime_base = os.path.join(
        os.getcwd(),
        '.kompos-runtime'
    )

    path = os.path.join(
        runtime_base,
        runner,
        kompos_config.root_path(runner),
    )

    return path


def validate_runner_version(kompos_config, runner):
    """
    Check if runner binary version is compatible with the
    version specified by the kompos configuration.
    """
    try:
        execution = Popen([runner, '--version'],
                          stdin=PIPE,
                          stdout=PIPE,
                          stderr=PIPE)
    except Exception:
        logging.exception(
            f"Runner {runner} does not appear to be installed, "
            f"please ensure terraform is in your PATH"
        )
        exit(1)

    expected_version = kompos_config.runner_version(runner)
    current_version, execution_error = execution.communicate()
    current_version = current_version.decode('utf-8').split('\n', 1)[0]

    if expected_version not in current_version:
        raise Exception(
            f"Runner [{runner}] should be {expected_version}, but you have {current_version}. "
            f"Please change your version."
        )

    return


def get_himl_args(args):
    parser = ConfigRunner.get_parser(argparse.ArgumentParser())

    # For config command, HIML args are already parsed directly into args
    # ConfigRunner().get_parser() adds all HIML args to the config subcommand
    if hasattr(args, 'command') and args.command == 'config':
        logger.debug("Using HIML arguments from config command")
        return args

    # For tfe command, HIML args are mixed with TFE-specific args
    # Return full args object to include both TFE and HIML args
    if hasattr(args, 'command') and args.command == 'tfe':
        logger.debug("Using HIML arguments from tfe command")
        return args

    # For helm command, return full args (subcommand, values-dir, dry-run, HIML args)
    if hasattr(args, 'command') and args.command == 'helm':
        logger.debug("Using HIML arguments from helm command")
        return args

    # For explore command, return full args for exploration options
    if hasattr(args, 'command') and args.command == 'explore':
        logger.debug("Using arguments from explore command")
        return args

    # For terraform/helmfile commands, use --himl flag if provided
    if hasattr(args, 'himl_args') and args.himl_args:
        himl_args = parser.parse_args(args.himl_args.split())
        logger.info(f"Extra himl arguments for {args.command}: {himl_args}")
        return himl_args

    # Default: return empty parsed args with defaults
    return parser.parse_args([])



"""
Console output helpers for consistent, colorful CLI output.
"""
import sys


# ANSI Color Codes
class Colors:
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[95m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    GRAY = '\033[90m'
    BOLD = '\033[1m'
    DIM = '\033[2m'
    RESET = '\033[0m'
    
    # Background colors
    BG_BLUE = '\033[44m'
    BG_CYAN = '\033[46m'


def print_error(message, details=None):
    """Print an error message in red."""
    print(f"{Colors.RED}ERROR: {message}{Colors.RESET}", file=sys.stderr)
    if details:
        if isinstance(details, list):
            for detail in details:
                print(detail, file=sys.stderr)
        else:
            print(details, file=sys.stderr)


def print_success(message):
    """Print a success message with checkmark."""
    print(f"{Colors.GREEN}✓{Colors.RESET} {message}")


def print_info(message, indent=0):
    """Print an info message with optional indentation."""
    prefix = "   " * indent
    print(f"{prefix}{message}")


def print_warning(message):
    """Print a warning message in yellow."""
    print(f"{Colors.YELLOW}⚠{Colors.RESET}  {message}")


def print_section_header(title, width=80):
    """Print a section header with separator lines."""
    print(f"\n{Colors.BOLD}{Colors.CYAN}{'═' * width}{Colors.RESET}")
    print(f"{Colors.BOLD}{Colors.CYAN}{title}{Colors.RESET}")
    print(f"{Colors.CYAN}{'═' * width}{Colors.RESET}")


def print_subsection(emoji, title, details=None):
    """Print a subsection with emoji and optional details."""
    print()
    print(f"{Colors.BOLD}{emoji} {title}{Colors.RESET}")
    if details:
        if isinstance(details, dict):
            for key, value in details.items():
                print(f"   {Colors.DIM}{key}:{Colors.RESET} {value}")
        elif isinstance(details, list):
            for item in details:
                print(f"   {item}")


def print_separator(width=80):
    """Print a separator line."""
    print(f"{Colors.CYAN}{'═' * width}{Colors.RESET}\n")


def print_summary(total_files=None, composition_files=None, config_files=None, elapsed_time=None, width=80):
    """Print a summary footer with metrics."""
    print(f"{Colors.CYAN}{'═' * width}{Colors.RESET}")
    
    summary_parts = []
    
    # Just show total files, no breakdown
    if total_files is not None:
        summary_parts.append(f"{total_files} files generated")
    
    # Time
    if elapsed_time is not None:
        if elapsed_time < 1:
            time_str = f"{elapsed_time * 1000:.0f}ms"
        else:
            time_str = f"{elapsed_time:.2f}s"
        summary_parts.append(f"{time_str}")
    
    if summary_parts:
        summary = f"{Colors.GREEN}✓{Colors.RESET} {Colors.BOLD}Completed{Colors.RESET}  {Colors.DIM}•{Colors.RESET}  " + f"  {Colors.DIM}•{Colors.RESET}  ".join(summary_parts)
        print(summary)
        print(f"{Colors.CYAN}{'═' * width}{Colors.RESET}\n")
    else:
        print()


def print_kvp(key, value, indent=1, dim_key=True):
    """Print a key-value pair with indentation."""
    prefix = "   " * indent
    if dim_key:
        print(f"{prefix}{Colors.DIM}{key}:{Colors.RESET} {value}")
    else:
        print(f"{prefix}{key}: {value}")


def format_config_path(path):
    """
    Format a hierarchical config path with colored keys and values.
    
    Example: configs/cloud=aws/project=demo/env=dev
    Returns: configs/cloud=aws/project=demo/env=dev with colors
    """
    if not path:
        return path
    
    parts = path.split('/')
    formatted_parts = []
    
    for part in parts:
        if '=' in part:
            key, value = part.split('=', 1)
            # Key in normal green, value in bold
            formatted_parts.append(f"{Colors.GREEN}{key}={Colors.RESET}{Colors.BOLD}{value}{Colors.RESET}")
        else:
            # No '=' means it's just a directory name
            formatted_parts.append(part)
    
    return '/'.join(formatted_parts)


# Specialized formatters for common Kompos operations

def print_composition_header(composition_name, composition_type=None, source=None, target=None, config_path=None):
    """Print a composition generation header with detailed info."""
    # Compact header with composition info
    if composition_type:
        title = f"TFE Composition: {Colors.GREEN}{Colors.BOLD}{composition_name}{Colors.RESET} {Colors.GREEN}({composition_type}){Colors.RESET}"
    else:
        title = f"TFE Composition: {Colors.GREEN}{Colors.BOLD}{composition_name}{Colors.RESET}"
    
    print()
    print(f"{Colors.CYAN}{'═' * 80}{Colors.RESET}")
    print(f"{Colors.BOLD}{title}")
    print(f"{Colors.CYAN}{'═' * 80}{Colors.RESET}")
    
    if config_path:
        formatted_path = format_config_path(config_path)
        print(f"   {Colors.DIM}Config:{Colors.RESET} {formatted_path}")
    
    # Paths section - will show success inline
    if source or target:
        print()
        if source:
            print(f"   {Colors.DIM}Source:{Colors.RESET} {source}")
        if target:
            print(f"   {Colors.DIM}Target:{Colors.RESET} {Colors.GREEN}{target}{Colors.RESET}")
    print()


def print_file_generation(file_type, output_path, format_type=None, size=None):
    """Print file generation info without format (compact)."""
    print(f"  • {file_type}: {output_path}")
    if size:
        print(f"    {Colors.DIM}Size:{Colors.RESET} {size}")



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os
from functools import reduce

import fastjsonschema
import yaml
from packaging.version import Version

from kompos import __version__

logger = logging.getLogger(__name__)

CONFIG_SCHEMA_PATH = "data/config_schema.json"


def get_value_or(dictionary, x_path, default=None):
    """
    Try to retrieve a value from a dictionary. Return the default if no such value is found.
    """
    keys = x_path.split("/")
    return reduce(
        lambda d, key: d.get(key, default)
        if isinstance(d, dict) else default, keys, dictionary)


class KomposConfig:
    """
    Parses all the available configuration files in order and merges them together.
    """

    DEFAULT_PATHS = [
        '/etc/kompos/.komposconfig.yaml',
        os.path.expanduser('~/.komposconfig.yaml'),
        os.path.join(os.getcwd(), '.komposconfig.yaml')
    ]

    def __init__(self, console_args, package_dir):
        self.config = dict()
        self.package_dir = package_dir
        self.validate = fastjsonschema.compile(self.read_schema())

        paths = self.DEFAULT_PATHS[:]

        parsed_files = []
        logger.debug(f"parsing {paths}")

        for config_path in paths:
            config_path = os.path.realpath(os.path.expanduser(config_path))
            if os.path.isfile(config_path):
                logger.debug(f"parsing {config_path}")
                with open(config_path) as f:
                    try:
                        config = yaml.safe_load(f.read())
                    except Exception as e:
                        logger.error(f"Failed to parse configuration file: {config_path}")
                        raise e

                    if isinstance(config, dict):
                        parsed_files.append(config_path)
                        self.config.update(config)
                    else:
                        logger.error(f"cannot parse yaml dict from file: {config_path}")

                try:
                    self.validate(config)
                except fastjsonschema.exceptions.JsonSchemaException as e:
                    logger.error(f"Schema validation failed for configuration file: {config_path}")
                    raise e

        self.parsed_files = parsed_files
        logger.debug(f"final kompos config: {self.config} from {parsed_files}")
        logger.debug(f"Loaded config from {', '.join([os.path.basename(f) for f in parsed_files])}")

    @property
    def kompos(self):
        """Quick access to komposconfig namespace"""
        return self.config.get('komposconfig', {})

    def read_schema(self):
        with open(os.path.join(self.package_dir, CONFIG_SCHEMA_PATH), "r") as f:
            return yaml.safe_load(f.read())

    def __contains__(self, item):
        return item in self.config

    def __getitem__(self, item):
        """
        Access komposconfig namespace items.
        All config must be under 'komposconfig' key.
        """
        if item not in self.kompos:
            raise KeyError(f"{item} not found in komposconfig namespace from {self.parsed_files}")

        return self.kompos[item]

    def vault_backend(self):
        if get_value_or(self.config, "vault/enabled"):
            os.environ["VAULT_ADDR"] = get_value_or(self.config, "vault/url")
            os.environ["VAULT_NAMESPACE"] = get_value_or(self.config, "vault/vault_namespace")
            os.environ["VAULT_USERNAME"] = get_value_or(self.config, "vault/svc_ldap_user")
            os.environ["VAULT_ROLE"] = get_value_or(self.config, "vault/svc_ldap_user_role")
            logger.info("Vault backend enabled")

    def validate_version(self):
        """Validate minimum required Kompos version"""
        min_kompos_version = self.get_kompos_setting("min_version")

        if not min_kompos_version:
            return

        if Version(__version__) < Version(min_kompos_version):
            raise Exception(
                f"The current kompos version '{__version__}' is lower than "
                f"the minimum required version '{min_kompos_version}'"
            )

    # ====================================================================================
    # Unified Configuration Access Methods
    # ====================================================================================
    def get_kompos_setting(self, path, default=None):
        """
        Get any setting from komposconfig namespace using dot-separated path.
        
        Args:
            path: Dot-separated path (e.g., 'compositions.properties.account.output_subdir')
            default: Default value if not found
        
        Returns:
            Value at path or default
            
        Examples:
            self.get_kompos_setting('terraform.version', 'latest')
            self.get_kompos_setting('compositions.properties.account.output_subdir', 'accounts')
        """
        return get_value_or(self.kompos, path.replace('.', '/'), default)

    def get_hierarchical_setting(self, runner, key, default=None):
        """
        Get setting with hierarchical lookup (DRY method for all hierarchical access).
        
        Lookup order:
        1. komposconfig.{runner}.{key}  (runner-specific)
        2. komposconfig.compositions.source.{key}  (for path-related keys)
        3. komposconfig.defaults.{key}  (global defaults)
        4. default parameter
        
        Args:
            runner: Runner name (e.g., 'terraform', 'tfe', 'helmfile')
            key: Setting key to lookup
            default: Default value if not found
        
        Returns:
            Setting value with hierarchical fallback
            
        Examples:
            self.get_hierarchical_setting('tfe', 'base_dir', './generated')
            self.get_hierarchical_setting('terraform', 'local_path', './compositions')
        """
        # 1. Check runner-specific setting
        if runner in self.kompos and key in self.kompos[runner]:
            value = self.kompos[runner][key]
            # Expand paths for local_path
            if key == 'local_path':
                return os.path.expanduser(value)
            return value

        # 2. Check composition source (for path-related keys)
        if key in ['local_path', 'root_path']:
            source = self.kompos.get('compositions', {}).get('source', {})
            if key in source:
                value = source[key]
                if key == 'local_path':
                    return os.path.expanduser(value)
                return value

        # 3. Check global defaults
        defaults = self.kompos.get('defaults', {})
        if key in defaults:
            return defaults[key]

        # 4. Return provided default
        return default

    # ====================================================================================
    # Composition Configuration Methods
    # ====================================================================================

    def excluded_config_keys(self, composition, default=[]):
        """Get excluded keys for a composition type"""
        return self.get_kompos_setting(f"compositions.config_keys.excluded.{composition}", default)

    def filtered_output_keys(self, composition, default=[]):
        """Get filtered keys for a composition type"""
        return self.get_kompos_setting(f"compositions.config_keys.filtered.{composition}", default)

    def composition_order(self, composition, default=[]):
        """Get execution order for compositions"""
        return self.get_kompos_setting(f"compositions.order.{composition}", default)

    # ====================================================================================
    # Runner Configuration Methods
    # ====================================================================================

    def runner_version(self, runner):
        """Get runner version with fallback to 'latest'"""
        return self.get_kompos_setting(f"{runner}.version", 'latest')

    def terraform_versioned_module_sources_enabled(self):
        """Check if Terraform versioned module sources feature is enabled"""
        return self.get_kompos_setting("terraform.versioned_module_sources", True)

    def repo_url(self, runner):
        """Get repository URL for a runner"""
        return self.kompos[runner]['repo']['url']

    def repo_name(self, runner):
        """Get repository name for a runner"""
        return self.kompos[runner]['repo']['name']

    def root_path(self, runner):
        """
        Get root_path with hierarchical lookup.
        Uses unified get_hierarchical_setting method.
        """
        return self.get_hierarchical_setting(runner, 'root_path', '')

    def local_path(self, runner):
        """
        Get local_path with hierarchical lookup.
        Uses unified get_hierarchical_setting method.
        """
        return self.get_hierarchical_setting(runner, 'local_path', './compositions')

    def get_runtime_setting(self, runner, key, default=None):
        """
        Get Kompos runtime setting from .komposconfig.yaml with hierarchical lookup.
        Uses unified get_hierarchical_setting method.
        
        IMPORTANT: This reads Kompos tool configuration (.komposconfig.yaml),
                   NOT layered/application config (configs/ directory).
        
        Runtime settings control Kompos behavior:
          - Where to find/generate files (paths, directories)
          - How to format output (yaml/json, extensions)
          - What to exclude (system keys, config keys)
          - Tool versions and feature flags
        
        Args:
            runner: Runner name (e.g., 'terraform', 'tfe')
            key: Setting key to lookup
            default: Default value if not found anywhere
        
        Returns:
            Runtime setting value from .komposconfig.yaml
            
        Examples:
            base_dir = kompos_config.get_runtime_setting('tfe', 'base_dir', './generated')
            fmt = kompos_config.get_runtime_setting('tfe', 'tfvars_format', 'yaml')
        """
        return self.get_hierarchical_setting(runner, key, default)

    # ====================================================================================
    # Composition Output & Naming Methods
    # ====================================================================================

    def get_composition_output_dir(self, base_dir, composition):
        """
        Get output directory for a composition with hierarchical lookup.
        
        Args:
            base_dir: Base directory for output
            composition: Composition name (e.g., 'cluster', 'account')
        
        Returns:
            Full path: {base_dir}/{output_subdir}/
        """
        output_subdir = self.get_kompos_setting(
            f"compositions.properties.{composition}.output_subdir",
            composition  # Fallback to composition name
        )
        return os.path.join(base_dir, output_subdir)

    def get_composition_name(self, raw_config, get_nested_value_fn):
        """
        Get the resolved composition instance identifier from layered config.
        
        Reads composition.instance which is defined with pure Himl interpolation:
          composition.instance: "{{account.name}}"  or  "{{cluster.fullName}}"
        
        Himl resolves the interpolation, returning the actual instance identifier:
        - account compositions: "my-account"
        - cluster compositions: "my-cluster-us-east-1"
        
        Args:
            raw_config: Himl-generated configuration dictionary
            get_nested_value_fn: Function to extract nested values from raw_config
            
        Returns:
            Resolved instance identifier (e.g., "my-account", "my-cluster-us-east-1")
        """
        import logging
        logger = logging.getLogger(__name__)

        # Read resolved composition instance from layered config (pure interpolation)
        name = get_nested_value_fn(raw_config, 'composition.instance')

        if name and '{{' not in str(name):
            logger.debug(f"Using composition instance from layered config: '{name}'")
            return name

        # Fallback if not defined or unresolved
        logger.warning("No resolved composition.instance in layered config")
        return None



# Copyright 2026 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os

from kompos.parser import SubParserConfig
from kompos.runners.terraform_helper import GenericTerraformRunner

logger = logging.getLogger(__name__)  

RUNNER_TYPE = "terraform"


class TerraformParser(SubParserConfig):
    def get_name(self):
        return RUNNER_TYPE

    def get_help(self):
        return 'Wrap common terraform tasks with full templated configuration support'

    def configure(self, parser):
        """Add terraform-specific arguments."""
        parser.add_argument('subcommand', help='One of the terraform commands', type=str)
        parser.add_argument('--dry-run', action='store_true',
                            help='Generate all files but do not execute terraform command')
        return parser

    def get_epilog(self):
        return '''
        Examples:
            # Run terraform plan
            kompos data/env=dev/region=va6/project=ee/cluster=experiments/composition=terraform terraform plan 
            # Run terraform apply on a specific composition
            kompos data/env=dev/region=va6/project=ee/cluster=experiments/composition=terraform/terraform=mycomposition terraform apply
        '''


class TerraformRunner(GenericTerraformRunner):
    """
    Local Terraform execution runner.
    
    Executes Terraform commands locally after generating configuration files.
    Uses temporary runtime directories for execution.
    """

    # Terraform subcommands that will need re-initialization
    SUBCMDS_WITH_INIT = ('plan', 'apply', 'destroy', 'import', 'state')

    # Terraform subcommands that use the variables file
    SUBCMDS_WITH_VARS = ('plan', 'apply', 'destroy', 'import')

    # Subcommand that reverses composition order
    REVERSE_COMPOSITION_CMD = 'destroy'

    # File naming conventions
    VERSIONED_EXTENSION = '.tf.versioned'
    CONFIG_FILENAME = 'variables.tfvars.json'
    CACHE_DIR = '~/.kompos/.terraform.d/plugin-cache'

    def __init__(self, kompos_config, config_path, execute):
        super(TerraformRunner, self).__init__(kompos_config, config_path, execute, RUNNER_TYPE)

    def run_configuration(self, args):
        self.ordered_compositions = True
        self.reverse = (self.REVERSE_COMPOSITION_CMD == args.subcommand)

    def execution_configuration(self, composition, config_path, default_output_path, raw_config,
                                filtered_keys, excluded_keys):
        # Source composition path
        source_composition_path = self.get_source_composition_path(composition, raw_config)

        # Runtime directory path (where we generate and execute)
        runtime_dir = self.get_runtime_dir(
            default_output_path,
            raw_config['cloud']['provider'],
            composition
        )

        # Clean and recreate runtime directory to ensure fresh state
        self.ensure_directory(runtime_dir, clean=True)
        logger.info(f'Using runtime directory: {runtime_dir}')

        # Note: provider.tf should use native TF variables, not generated from Hiera

        # Generate variables.tfvars.json
        variables_path = os.path.join(runtime_dir, self.CONFIG_FILENAME)
        self.generate_terraform_config(
            target_file=variables_path,
            config_path=config_path,
            filtered_keys=filtered_keys,
            excluded_keys=excluded_keys + self.TERRAFORM_SYSTEM_KEYS,
            output_format='json',
            enclosing_key='config',
            print_data=True
        )

        # Process composition files (versioned + static)
        if self.is_versioned_module_sources_enabled():
            self.process_composition_files(source_composition_path, runtime_dir, raw_config)

    @staticmethod
    def execution(args, extra_args, default_output_path, composition, raw_config):
        # Runtime directory for TF execution
        runtime_dir = os.path.join(default_output_path, raw_config['cloud']['provider'], composition)

        # Handle dry-run mode: skip terraform execution
        if args.dry_run:
            logger.info(f'DRY-RUN: Skipping terraform execution for composition: {composition}')
            logger.info(f'DRY-RUN: Generated files are in: {runtime_dir}')
            # Return a no-op command that always succeeds
            return dict(command="echo 'Dry-run mode: skipping terraform execution'")

        var_file = f'-var-file="{TerraformRunner.CONFIG_FILENAME}"' if args.subcommand in TerraformRunner.SUBCMDS_WITH_VARS else ''
        terraform_env_config = f'export TF_PLUGIN_CACHE_DIR="{TerraformRunner.local_config_dir()}"'

        remove_cache = TerraformRunner.remove_local_cache_cmd(args.subcommand)
        extra_args_str = ' '.join(extra_args)
        cmd = (f"cd {runtime_dir} && "
               f"{remove_cache} "
               f"{terraform_env_config} ; terraform init && terraform {args.subcommand} {var_file} {extra_args_str}")

        return dict(command=cmd)

    @staticmethod
    def remove_local_cache_cmd(subcommand):
        """Generate command to remove local .terraform cache if needed."""
        if subcommand in TerraformRunner.SUBCMDS_WITH_INIT:
            return 'rm -rf .terraform &&'
        return ''

    @staticmethod
    def local_config_dir(directory=None):
        """
        Create and return path to Terraform plugin cache directory.
        
        Args:
            directory: Optional cache directory path (default: TerraformRunner.CACHE_DIR)
            
        Returns:
            str: Expanded path to cache directory
        """
        if directory is None:
            directory = TerraformRunner.CACHE_DIR

        try:
            expanded_path = os.path.expanduser(directory)
            os.makedirs(expanded_path, exist_ok=True)
            return expanded_path
        except IOError:
            logging.error(f"Failed to create dir in path: {directory}")



# Copyright 2026 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os
import time

from kompos.parser import SubParserConfig
from kompos.runners.terraform_helper import GenericTerraformRunner
from kompos.helpers import console

logger = logging.getLogger(__name__)

RUNNER_TYPE = "tfe"


class TFEParserConfig(SubParserConfig):
    def get_name(self):
        return RUNNER_TYPE

    def get_help(self):
        return 'Generate Terraform Enterprise (TFE) workspace and tfvars configurations'

    def configure(self, parser):
        """Add TFE-specific arguments and standard HIML arguments."""
        # TFE subcommand
        parser.add_argument('subcommand',
                            metavar='SUBCOMMAND',
                            choices=['generate'],
                            help='Action to perform: generate')

        # TFE-specific options
        parser.add_argument('--tfvars-only',
                            action='store_true',
                            help='Generate only tfvars file (skip workspace config)')
        parser.add_argument('--workspace-only',
                            action='store_true',
                            help='Generate only workspace config (skip tfvars)')

        # Add HIML arguments
        self.add_himl_arguments(parser)
        return parser

    def get_epilog(self):
        return '''
Examples:
  # Generate TFE workspace and tfvars for an account composition
  kompos configs/cloud=aws/project=demo/env=dev/composition=account tfe generate
  
  # Generate for a cluster composition
  kompos configs/cloud=aws/project=demo/env=dev/region=us-west-2/cluster=demo01/composition=cluster tfe generate
  
  # Generate only tfvars (skip workspace config)
  kompos configs/.../composition=cluster tfe generate --tfvars-only
  
  # Generate with custom filters
  kompos configs/.../composition=cluster tfe generate --filter cluster --filter vpc --exclude node_groups

Output:
  - Composition files: generated/{composition_type}/{instance}/
  - Workspace config:  generated/workspaces/{instance}.workspace.yaml
  - Tfvars file:       generated/{composition_type}/{instance}/generated.tfvars.yaml

For more information, see: docs/GUIDE.md
        '''


class TFERunner(GenericTerraformRunner):
    def __init__(self, kompos_config, config_path, execute):
        super(TFERunner, self).__init__(kompos_config, config_path, execute, RUNNER_TYPE)

        # Cache TFE-specific configuration from .komposconfig.yaml for performance
        # base_dir and use_composition_output_dir are inherited from GenericTerraformRunner
        
        # Generation config
        self.generate_compositions = self.kompos_config.get_runtime_setting(
            self.runner_type, 'generation_config.generate_compositions', True)

        # Workspace configuration
        workspace_config = self.kompos_config.get_runtime_setting(
            self.runner_type, 'workspaces_config', {})
        
        self.generate_workspaces = workspace_config.get('generate', True)
        self.workspace_config_key = workspace_config.get('config_key', 'workspace')
        
        # Build workspaces directory from base + subdir
        base_output_dir = self.kompos_config.get_runtime_setting(
            self.runner_type, 'generation_config.base_output_dir', './generated')
        workspaces_subdir = workspace_config.get('workspaces_sub_dir', 'workspaces')
        self.workspaces_dir = os.path.join(base_output_dir, workspaces_subdir)
        
        self.workspace_extension = workspace_config.get('workspace_extension', '.workspace.yaml')
        self.workspace_format = self.extract_format_from_extension(self.workspace_extension)

        # Tfvars configuration
        tfvars_config = self.kompos_config.get_runtime_setting(
            self.runner_type, 'tfvars_config', {})
        
        self.tfvars_extension = tfvars_config.get('extension', '.tfvars.yaml')
        self.tfvars_format = self.extract_format_from_extension(self.tfvars_extension)
        
        # If tfvars_filename is None/not set, will use composition name (backward compatible)
        # If set to a string like "generated", will use that instead
        self.tfvars_filename = tfvars_config.get('filename', None)

    def run_configuration(self, args):
        self.validate_runner = False
        self.ordered_compositions = False
        self.reverse = False
        self.generate_output = False

    def execution_configuration(self, composition, config_path, default_output_path, raw_config,
                                filtered_keys, excluded_keys):
        args = self.himl_args
        
        # Start timing
        start_time = time.time()
        composition_files = 0
        config_files = 0

        # Determine what to generate
        generate_tfvars = not args.workspace_only
        generate_workspace = not args.tfvars_only and self.generate_workspaces
        generate_composition = self.generate_compositions

        # Generate per-cluster composition (working_directory for TFE)
        if generate_composition:
            file_count = self.generate_composition(composition, raw_config)
            composition_files = file_count if file_count else 0
            # Add separator after composition files if we're generating config files
            if generate_tfvars or generate_workspace:
                print()

        # Generate tfvars file
        if generate_tfvars:
            self.generate_tfvars(composition, config_path, raw_config, filtered_keys, excluded_keys)
            config_files += 1

        # Generate workspace config file
        if generate_workspace:
            self.generate_workspace_config(config_path, raw_config)
            config_files += 1
        
        # Print summary with metrics
        total_files = composition_files + config_files
        elapsed_time = time.time() - start_time
        console.print_summary(total_files=total_files, elapsed_time=elapsed_time)

    def generate_composition(self, composition, raw_config):
        """
        Generate per-cluster TFE working_directory with modules.
        
        Processes .tf.versioned files and copies static .tf files.
        Note: Provider config should use native TF variables, not generated from Hiera.
        
        Args:
            composition: Name of the composition (e.g., 'cluster')
            raw_config: Configuration dictionary from hiera
            
        Returns:
            int: Total number of files processed
        """
        cluster_name = self.get_composition_name(raw_config)
        if not cluster_name:
            return 0

        # Source composition path
        source_composition_path = self.get_source_composition_path(composition, raw_config)

        if not os.path.exists(source_composition_path):
            raise FileNotFoundError(
                f'Source composition directory does not exist: {source_composition_path}\n'
                f'Composition: {composition}, Cloud provider: {raw_config["cloud"]["provider"]}'
            )

        # Target composition path: base/composition_type/composition_instance
        # - base: ./generated (from .komposconfig.yaml defaults.base_dir)
        # - composition_type: accounts/ or clusters/ (from .komposconfig.yaml properties)
        # - composition_instance: my-account/ (from composition.output_dir interpolation)
        composition_output_dir = self.kompos_config.get_composition_output_dir(self.base_dir, composition)
        instance_dir = self.get_composition_output_dir(raw_config)
        target_dir = self.build_output_path(
            base_dir=composition_output_dir,
            name=instance_dir,
            use_subdir=self.use_composition_subdir
        )

        self.ensure_directory(target_dir)

        console.print_composition_header(
            composition_name=cluster_name,
            composition_type=composition,
            source=source_composition_path,
            target=target_dir,
            config_path=self.config_path
        )

        # Process all composition files (versioned + static)
        # Note: provider.tf should use native TF variables, not generated from Hiera
        file_count = 0
        if self.is_versioned_module_sources_enabled():
            processed, copied = self.process_composition_files(source_composition_path, target_dir, raw_config)
            file_count = processed + copied

        print()  # Blank line for spacing
        console.print_success(f"Composition files copied ({file_count} files)")
        
        return file_count

    def generate_tfvars(self, composition, config_path, raw_config, filtered_keys, excluded_keys):
        """Generate the tfvars file for TFE (per-composition)"""
        workspace_name = self.get_composition_name(raw_config)
        if not workspace_name:
            return

        # Determine base filename: use configured name OR composition name for backward compatibility
        base_filename = self.tfvars_filename if self.tfvars_filename else workspace_name
        tfvars_filename = f'{base_filename}{self.tfvars_extension}'

        # Build output path: base/composition_type/composition_instance/filename
        composition_output_dir = self.kompos_config.get_composition_output_dir(self.base_dir, composition)
        instance_dir = self.get_composition_output_dir(raw_config)
        output_file = self.build_output_path(
            base_dir=composition_output_dir,
            name=instance_dir,
            filename=tfvars_filename,
            use_subdir=self.use_composition_subdir
        )

        # Ensure output directory exists
        self.ensure_directory(output_file, is_file_path=True)

        self.generate_terraform_config(
            target_file=output_file,
            config_path=config_path,
            filtered_keys=filtered_keys,
            excluded_keys=excluded_keys + self.TFE_SYSTEM_KEYS + self.TERRAFORM_SYSTEM_KEYS,
            output_format=self.tfvars_format,
            enclosing_key='config',
            print_data=False
        )
        
        print()  # Blank line for spacing
        console.print_success("Configuration generated")
        console.print_file_generation("tfvars", output_file)

    def generate_workspace_config(self, config_path, raw_config):
        """Generate the workspace configuration file for TFE workspace creation"""
        workspace_name = self.get_composition_name(raw_config)
        if not workspace_name:
            return

        # Build output file path
        output_file = self.build_output_path(
            base_dir=self.workspaces_dir,
            filename=f'{workspace_name}{self.workspace_extension}'
        )

        # Ensure output directory exists
        self.ensure_directory(output_file, is_file_path=True)

        # Filter to workspace key (config_key from .komposconfig.yaml)
        self.generate_terraform_config(
            target_file=output_file,
            config_path=config_path,
            filtered_keys=[self.workspace_config_key],
            output_format=self.workspace_format,
            print_data=False
        )
        
        console.print_file_generation("workspace", output_file)

    @staticmethod
    def execution(args, extra_args, default_output_path, composition, raw_config):
        """No actual execution needed - files are generated in execution_configuration"""
        cmd = ""
        return dict(command=cmd)



# Copyright 2026 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import json
import logging
import os
import sys
import time

import yaml
from deepmerge import Merger

from kompos.parser import SubParserConfig
from kompos.runner import GenericRunner
from kompos.helpers import console

logger = logging.getLogger(__name__)

RUNNER_TYPE = "helm"

# Key injected into context during interpolation, stripped from output.
# Isolates chart values from kompos context keys — never appears in output files.
DEFAULT_ENCLOSING_KEY = "__helm_values__"

# Deepmerge strategy: dicts merge, lists append, conflicts override
_MERGER = Merger(
    [(dict, ["merge"]), (list, ["append"])],
    ["override"],
    ["override"]
)


class HelmParserConfig(SubParserConfig):
    def get_name(self):
        return RUNNER_TYPE

    def get_help(self):
        return 'Render himl-templated Helm values for a cluster'

    def configure(self, parser):
        parser.add_argument('subcommand',
                            metavar='SUBCOMMAND',
                            choices=['generate', 'list'],
                            help='Action to perform: generate | list')

        # generate options
        parser.add_argument('--charts-dir',
                            metavar='PATH',
                            help='Root dir of chart subdirs; defaults to helm.config.charts_dir in .komposconfig.yaml')
        parser.add_argument('--chart-dir',
                            metavar='PATH',
                            help='Single chart directory for local dev (chart name inferred from dir name)')
        parser.add_argument('--dry-run',
                            action='store_true',
                            help='Print rendered output to stdout, do not write files')

        # list options
        parser.add_argument('--output-format',
                            dest='list_format',
                            choices=['table', 'yaml', 'json'],
                            default='table',
                            help='Output format for list (default: table)')

        self.add_himl_arguments(parser)
        return parser

    def get_epilog(self):
        return '''
Examples:
  # Render all enabled charts (charts_dir from .komposconfig.yaml)
  kompos configs/cloud=aws/.../cluster=cheeta/composition=helm-values helm generate

  # Override charts directory
  kompos configs/.../composition=helm-values \\
      helm generate --charts-dir /path/to/k8s-apps-deploy/applications/laser

  # Single chart (local dev)
  kompos configs/.../composition=helm-values \\
      helm generate --chart-dir /path/to/applications/laser/contour-ingress-controller

  # Dry-run (template under development)
  kompos configs/.../composition=helm-values \\
      helm generate --dry-run --skip-interpolation-validation

  # List enabled charts
  kompos configs/.../composition=helm-values helm list
        '''


class HelmRunner(GenericRunner):
    """
    Renders himl-templated Helm values for a cluster.

    Context assembly:
      1. raw_config  — full kompos hierarchy (cluster, env, region, account, ...)
      2. tfe_outputs — runtime infra state (global.infra.* from tfe-outputs.yaml)

    Values generation (per enabled chart):
      context[enclosing_key] = values_dict     # inject under internal key
      resolve_interpolations(context)           # resolve {{}} in-place
      rendered = context.pop(enclosing_key)     # extract clean values

    Output:
      generated/clusters/{cluster}/argoapps/{app}.yaml
    """


    def __init__(self, kompos_config, config_path, execute):
        super(HelmRunner, self).__init__(kompos_config, config_path, execute, RUNNER_TYPE)

        helm_config = self.kompos_config.get_runtime_setting(
            self.runner_type, 'config', {})

        self.enclosing_key      = helm_config.get('enclosing_key',    DEFAULT_ENCLOSING_KEY)
        self.tfe_outputs_path   = helm_config.get('tfe_outputs_path', 'outputs/tfe-outputs.yaml')
        self.argoapps_subdir    = helm_config.get('output_subdir',    'argoapps')
        self.base_output_dir    = helm_config.get('base_output_dir',  './generated')
        self.default_charts_dir = helm_config.get('charts_dir',       None)

    def run_configuration(self, args):
        self.validate_runner = False
        self.ordered_compositions = False
        self.reverse = False
        self.generate_output = False

    def execution_configuration(self, composition, config_path, default_output_path,
                                raw_config, filtered_keys, excluded_keys):
        args = self.himl_args

        if args.subcommand == 'list':
            self.run_list(raw_config, getattr(args, 'list_format', 'table'))
            return

        if args.subcommand == 'generate':
            charts_dir = args.charts_dir or self.default_charts_dir
            if not charts_dir and not args.chart_dir:
                console.print_error(
                    "Missing charts directory for 'generate'.",
                    "Provide --charts-dir PATH, --chart-dir PATH, "
                    "or set helm.config.charts_dir in .komposconfig.yaml"
                )
                sys.exit(1)
            self.run_generate(args, raw_config, charts_dir)
            return

    @staticmethod
    def execution(args, extra_args, default_output_path, composition, raw_config):
        return dict(command="")

    # ── list ──────────────────────────────────────────────────────────────────

    def run_list(self, raw_config, fmt):
        cluster_name = self.get_composition_name(raw_config)
        charts = self.enabled_charts(raw_config)

        print(f"\nEnabled charts for cluster: {cluster_name}\n")

        if fmt == 'yaml':
            print(yaml.dump({'helm': {'charts': {k: v for k, v in charts.items()}}},
                            default_flow_style=False))
            return

        if fmt == 'json':
            print(json.dumps({'helm': {'charts': {k: v for k, v in charts.items()}}}, indent=2))
            return

        if not charts:
            print("  (no charts enabled)")
            return

        print(f"  {'CHART':<45} {'VERSION':<12} STATUS")
        print(f"  {'─'*45} {'─'*12} ──────")
        for name, cfg in sorted(charts.items()):
            version = cfg.get('version', '(unset)')
            print(f"  {name:<45} {version:<12} enabled")
        print()

    # ── generate ──────────────────────────────────────────────────────────────

    def run_generate(self, args, raw_config, charts_dir):
        start = time.time()
        cluster_name = self.get_composition_name(raw_config)

        if not cluster_name:
            console.print_error(
                "Could not determine cluster name.",
                "Ensure composition.instance is set in composition=helm-values config."
            )
            sys.exit(1)

        console.print_section_header(f"Helm Values Render: {cluster_name}")

        context = self.build_context(raw_config, cluster_name)

        # --chart-dir: single chart mode — path comes directly from the arg
        if args.chart_dir:
            chart_dir_abs = os.path.abspath(args.chart_dir)
            app_name      = os.path.basename(chart_dir_abs)
            chart_files   = {app_name: os.path.join(chart_dir_abs, 'values.yaml')}
            disabled, untracked = set(), set()

        # --charts-dir / komposconfig default: scan filesystem, filter to enabled
        else:
            charts_dir    = os.path.abspath(charts_dir)
            charts_on_fs  = self.find_charts(charts_dir)
            enabled, disabled, untracked = self.categorize_charts(charts_on_fs, raw_config)

            if not enabled:
                console.print_warning("No enabled charts found in helm.charts.*")
                return

            chart_files = {
                app_name: os.path.join(charts_dir, app_name, 'values.yaml')
                for app_name in enabled
            }

        argoapps_dir = os.path.join(
            self.base_output_dir, 'clusters', cluster_name, self.argoapps_subdir)

        dry_run = getattr(args, 'dry_run', False)
        rendered_count = 0

        for app_name, values_path in sorted(chart_files.items()):
            rendered = self.render_values(app_name, values_path, context)
            if rendered is None:
                continue

            if dry_run:
                print(f"\n# ── {app_name} ──────────────────────────────────────")
                print(yaml.dump(rendered, default_flow_style=False))
            else:
                output_file = os.path.join(argoapps_dir, f"{app_name}.yaml")
                self.ensure_directory(output_file, is_file_path=True)
                with open(output_file, 'w') as f:
                    yaml.dump(rendered, f, default_flow_style=False, sort_keys=False)
                console.print_success(f"Rendered {app_name}")
                console.print_file_generation("argoapps", output_file)

            rendered_count += 1

        if not dry_run:
            self.prune_argoapps(argoapps_dir, rendered=set(chart_files.keys()))

        self.report_chart_status(disabled, untracked)
        console.print_summary(total_files=rendered_count, elapsed_time=time.time() - start)

    # ── context ───────────────────────────────────────────────────────────────

    def build_context(self, raw_config, cluster_name):
        """
        Merge kompos hierarchy + TFE outputs into the interpolation context.

        raw_config  → top-level keys from the kompos tree walk
        tfe_outputs → global.infra.* overlaid on top
        """
        context = dict(raw_config)

        tfe_path = os.path.join(
            self.base_output_dir, 'clusters', cluster_name, self.tfe_outputs_path)

        if not os.path.exists(tfe_path):
            console.print_error(
                f"TFE outputs not found: {tfe_path}",
                "Ensure terraform has been applied and outputs exported before rendering helm values."
            )
            sys.exit(1)

        with open(tfe_path) as f:
            tfe_outputs = yaml.safe_load(f) or {}
        context = _MERGER.merge(context, tfe_outputs)
        logger.info(f"Loaded TFE outputs: {tfe_path}")

        context.pop(self.enclosing_key, None)
        return context

    # ── chart discovery ───────────────────────────────────────────────────────

    def find_charts(self, charts_dir):
        """Scan filesystem: return names of all chart dirs in charts_dir that contain values.yaml."""
        if not charts_dir or not os.path.isdir(charts_dir):
            return set()
        return {
            d for d in os.listdir(charts_dir)
            if os.path.isfile(os.path.join(charts_dir, d, 'values.yaml'))
        }

    def charts_config(self, raw_config):
        """Return the raw helm.charts.* dict from the kompos hierarchy."""
        return raw_config.get('helm', {}).get('charts', {})

    def categorize_charts(self, charts_on_fs, raw_config):
        """
        Compare filesystem state with hierarchy config.

        Returns three disjoint sets:
          enabled   — in helm.charts.* with enabled: true AND present on disk
          disabled  — in helm.charts.* with enabled: false AND present on disk
          untracked — present on disk but not in helm.charts.* at all
        """
        all_charts = self.charts_config(raw_config)
        configured = set(all_charts.keys())

        enabled   = {name for name, cfg in all_charts.items()
                     if isinstance(cfg, dict) and cfg.get('enabled', True)
                     and name in charts_on_fs}
        disabled  = {name for name, cfg in all_charts.items()
                     if isinstance(cfg, dict) and not cfg.get('enabled', True)
                     and name in charts_on_fs}
        untracked = charts_on_fs - configured

        return enabled, disabled, untracked

    def enabled_charts(self, raw_config):
        """Return charts enabled in helm.charts.* (used by run_list)."""
        charts_data = self.charts_config(raw_config)
        return {
            name: cfg
            for name, cfg in (charts_data or {}).items()
            if isinstance(cfg, dict) and cfg.get('enabled', True)
        }

    # ── rendering ─────────────────────────────────────────────────────────────

    def render_values(self, app_name, values_path, context):
        """
        Render a single values.yaml template against the interpolation context.

        Inject values under enclosing_key → resolve {{}} → pop and return clean values.
        Context is modified in-place but restored (enclosing_key is always popped).
        """
        if not os.path.exists(values_path):
            console.print_warning(f"values.yaml not found for '{app_name}': {values_path}")
            return None

        with open(values_path) as f:
            template = yaml.safe_load(f)

        if template is None:
            console.print_warning(f"Empty values.yaml for '{app_name}': {values_path}")
            return None

        if not isinstance(template, dict):
            console.print_warning(
                f"values.yaml for '{app_name}' is not a YAML mapping — skipping "
                f"(got {type(template).__name__})"
            )
            return None

        context[self.enclosing_key] = template

        try:
            self.resolve_interpolations(context)
        except Exception as e:
            logger.error(f"Interpolation failed for '{app_name}': {e}")
            context.pop(self.enclosing_key, None)
            raise

        return context.pop(self.enclosing_key)

    # ── reporting ─────────────────────────────────────────────────────────────

    def prune_argoapps(self, argoapps_dir, rendered):
        """
        Remove argoapps/ files for charts no longer rendered.

        argoapps/ is fully owned by helm generate — any .yaml not in the
        current rendered set is stale (disabled or removed from helm.charts.*).
        Pruning ensures ArgoCD never applies values from a disabled chart.
        """
        if not os.path.isdir(argoapps_dir):
            return
        for filename in os.listdir(argoapps_dir):
            if not filename.endswith('.yaml'):
                continue
            chart_name = filename[:-5]   # strip .yaml
            if chart_name not in rendered:
                stale_path = os.path.join(argoapps_dir, filename)
                os.remove(stale_path)
                console.print_warning(f"Pruned stale argoapps file: {stale_path}")

    def report_chart_status(self, disabled, untracked):
        """Report disabled and untracked charts after categorize_charts comparison."""
        disabled  = sorted(disabled)
        untracked = sorted(untracked)

        if disabled:
            print(f"\n  Disabled:")
            for name in disabled:
                print(f"    {name}")

        if untracked:
            print(f"\n  Untracked:")
            for name in untracked:
                print(f"    {name}")

        if disabled or untracked:
            print()



# Copyright 2026 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Generic Terraform runner base class and helper utilities.

Provides:
- Base class for Terraform and TFE runners
- Module version pinning via .tf.versioned files
- Composition processing (versioned + static files)
"""

import logging
import os
import re
import shutil

from kompos.runner import GenericRunner
from kompos.helpers import console

logger = logging.getLogger(__name__)


class TerraformVersionedSourceProcessor:
    """
    Processes .tf.versioned files with module source version interpolation.
    
    This processor enables per-environment/cluster module version pinning by:
    1. Finding .tf.versioned files in composition directories
    2. Interpolating {{key.path}} placeholders with config values
    3. Generating static .tf files for Terraform to use
    4. Copying other static files as-is
    
    Example:
        main.tf.versioned:
            module "vpc" {
              source = "git::https://github.com/org/repo.git?ref={{vpc.module_version}}"
            }
        
        Config: {"vpc": {"module_version": "v2.0.0"}}
        
        Generates main.tf:
            module "vpc" {
              source = "git::https://github.com/org/repo.git?ref=v2.0.0"
            }
    """
    
    @staticmethod
    def process(source_dir, target_dir, config, versioned_extension=".tf.versioned"):
        """
        Process all terraform files in source directory and generate in target directory.
        
        Single loop through files:
        - .tf.versioned files → process and generate .tf files
        - Other files → copy as-is
        
        Args:
            source_dir: Path to the source composition directory
            target_dir: Path to the runtime directory
            config: Configuration dictionary for value interpolation
            versioned_extension: File extension for versioned terraform files 
                               (default: .tf.versioned, configurable via .komposconfig.yaml)
        
        Returns:
            Tuple of (processed_count, copied_count)
        """
        if not os.path.exists(source_dir):
            logger.warning(f'Source directory does not exist: {source_dir}')
            return 0, 0

        processed_count = 0
        copied_count = 0

        for filename in os.listdir(source_dir):
            source_file = os.path.join(source_dir, filename)

            if os.path.isdir(source_file):
                continue

            # Process .tf.versioned files
            if TerraformVersionedSourceProcessor._is_versioned_file(filename, versioned_extension):
                TerraformVersionedSourceProcessor._process_file(
                    source_file, target_dir, config, versioned_extension)
                processed_count += 1

            # Skip files generated from hiera
            elif TerraformVersionedSourceProcessor._should_skip_file(filename):
                continue

            # Skip .tf files that have a .versioned counterpart
            elif filename.endswith('.tf') and TerraformVersionedSourceProcessor._has_versioned_counterpart(
                    source_dir, filename, versioned_extension):
                logger.debug(f'Skipping {filename} (generated from .versioned)')
                continue

            # Copy everything else
            else:
                TerraformVersionedSourceProcessor._copy_file(source_file, target_dir, filename)
                copied_count += 1

        if processed_count > 0:
            logger.debug('Processed %d versioned file(s)', processed_count)
        if copied_count > 0:
            logger.debug('Copied %d static file(s)', copied_count)
        
        return processed_count, copied_count

    @staticmethod
    def _is_versioned_file(filename, versioned_extension):
        """Check if file is a .tf.versioned template."""
        return filename.endswith(versioned_extension)

    @staticmethod
    def _should_skip_file(filename):
        """
        Check if file should be skipped during processing.
        
        Currently no files are skipped - all .tf files are copied.
        This method exists for future extensibility.
        """
        return False

    @staticmethod
    def _has_versioned_counterpart(source_dir, filename, versioned_extension):
        """Check if a .tf file has a corresponding .tf.versioned file."""
        versioned_file = os.path.join(source_dir, filename.replace('.tf', versioned_extension))
        return os.path.exists(versioned_file)

    @staticmethod
    def _copy_file(source_file, target_dir, filename):
        """Copy a file to target directory."""
        shutil.copy2(source_file, os.path.join(target_dir, filename))
        logger.debug(f'Copied {filename}')

    @staticmethod
    def _process_file(versioned_file, target_dir, config, versioned_extension):
        """
        Process a single .tf.versioned file.
        
        Args:
            versioned_file: Path to the .tf.versioned template file
            target_dir: Directory where the generated .tf file will be written
            config: Configuration dictionary for value interpolation
            versioned_extension: The extension to replace (e.g., '.tf.versioned')
        """
        # Read the versioned file
        with open(versioned_file, 'r') as f:
            content = f.read()

        # Interpolate {{key.path}} placeholders and get interpolation info
        interpolated_content, interpolations = TerraformVersionedSourceProcessor.interpolate_sources(content, config)

        # Write to .tf file (remove .versioned extension) in target directory
        output_filename = os.path.basename(versioned_file).replace(versioned_extension, '.tf')
        output_file = os.path.join(target_dir, output_filename)

        with open(output_file, 'w') as f:
            f.write(interpolated_content)

        # Display interpolations to user
        if interpolations:
            for placeholder, value in interpolations:
                # Extract module name from placeholder (e.g., "vpc.module_version" -> "vpc")
                module_name = placeholder.split('.')[0] if '.' in placeholder else placeholder
                console.print_info(f"  → versioned: {module_name} = {value}", indent=1)
        
        logger.debug(f'Generated {output_filename} from {os.path.basename(versioned_file)}')

    @staticmethod
    def interpolate_sources(content, config):
        """
        Interpolate {{key.path}} patterns in module source lines only.
        
        Why custom interpolation instead of HIML?
        - HIML only processes YAML files
        - These are Terraform .tf files (HCL syntax)
        - We only interpolate 'source' lines for safety (avoid breaking Terraform syntax)
        
        This method processes terraform configuration line by line, looking for
        module source declarations containing placeholders. Only lines with both
        'source' and '{{...}}' are processed to ensure we don't accidentally
        interpolate other parts of the configuration.
        
        Args:
            content: The terraform file content as string
            config: Configuration dictionary for value lookup
            
        Returns:
            Tuple of (interpolated_content, interpolations)
            - interpolated_content: String with interpolated values
            - interpolations: List of (placeholder, value) tuples for display
            
        Example:
            Input:
              source = "git::...?ref={{vpc.module_version}}"
            
            Config: {"vpc": {"module_version": "v2.0.0"}}
            
            Output:
              source = "git::...?ref=v2.0.0"
        """
        lines = content.split('\n')
        output_lines = []
        interpolations = []

        for line in lines:
            # Only interpolate lines containing 'source' and placeholders
            if 'source' in line and '{{' in line:
                # Find all {{key.path}} patterns
                placeholders = re.findall(r'\{\{([^}]+)\}\}', line)

                for placeholder in placeholders:
                    # Get value from config using GenericRunner's static helper
                    value = GenericRunner.get_nested_value(config, placeholder)

                    if value is None:
                        raise ValueError(
                            f'Config key "{placeholder}" not found for interpolation.\n'
                            f'Line: {line.strip()}\n'
                            f'Required config keys must be present when using versioned sources.\n'
                            f'Add "{placeholder}" to your hierarchical configuration.'
                        )

                    # Replace placeholder with value
                    line = line.replace(f'{{{{{placeholder}}}}}', str(value))
                    interpolations.append((placeholder, str(value)))
                    logger.debug(f'Interpolated {{{{{placeholder}}}}} -> {value}')

            output_lines.append(line)

        return '\n'.join(output_lines), interpolations


class GenericTerraformRunner(GenericRunner):
    """
    Base class for Terraform and TFE runners.
    
    Provides shared functionality:
    - Versioned module source processing
    - Composition file handling
    - Path management utilities
    """

    def __init__(self, kompos_config, config_path, execute, runner_type):
        super(GenericTerraformRunner, self).__init__(kompos_config, config_path, execute, runner_type)

        # Get versioned extension from config (default: .tf.versioned)
        self.versioned_extension = self.kompos_config.get_runtime_setting(
            self.runner_type, 'versioned_extension', '.tf.versioned')

        # Load system exclusion keys from config (required in .komposconfig.yaml)
        system_keys_config = self.kompos_config.kompos.get('compositions', {}).get('system_keys', {})

        # Always exclude 'komposconfig' key from output (Kompos runtime settings)
        self.TERRAFORM_SYSTEM_KEYS = ['komposconfig'] + system_keys_config.get('terraform', [])
        self.TFE_SYSTEM_KEYS = ['komposconfig'] + system_keys_config.get('tfe', [])

        # Load hierarchical config values from .komposconfig.yaml
        self.base_dir = self.kompos_config.get_runtime_setting(self.runner_type, 'base_dir', './generated')
        self.use_composition_subdir = self.kompos_config.get_runtime_setting(self.runner_type, 'use_composition_subdir',
                                                                             True)

    def get_source_composition_path(self, composition, raw_config):
        """
        Get path to source composition templates.
        
        Args:
            composition: Composition name (e.g., 'cluster')
            raw_config: Configuration dictionary with cloud type
            
        Returns:
            Path to source composition directory
        """
        return os.path.join(
            self.kompos_config.local_path(self.runner_type),
            self.kompos_config.root_path(self.runner_type),
            raw_config["cloud"]["provider"],
            composition
        )

    def process_composition_files(self, source_dir, target_dir, raw_config):
        """
        Process all composition files (versioned + static).
        
        Handles:
        - .tf.versioned files → process and generate .tf
        - Static .tf files → copy as-is (unless they have .versioned counterpart)
        - Other files → copy as-is
        
        Args:
            source_dir: Source composition directory
            target_dir: Target directory for generated files
            raw_config: Configuration for interpolation
            
        Returns:
            tuple: (processed_count, copied_count)
        """
        if not os.path.exists(source_dir):
            logger.warning(f'Source composition directory does not exist: {source_dir}')
            return 0, 0

        return TerraformVersionedSourceProcessor.process(
            source_dir, target_dir, raw_config, self.versioned_extension)

    def is_versioned_module_sources_enabled(self):
        """Check if versioned module sources feature is enabled."""
        return self.kompos_config.terraform_versioned_module_sources_enabled()

    def build_output_path(self, base_dir, name=None, filename=None, use_subdir=True):
        """
        Build output path with optional subdirectory structure.
        
        Provides consistent path construction across runners with configurable
        subdirectory nesting.
        
        Args:
            base_dir: Base output directory
            name: Optional name for subdirectory (e.g., cluster name)
            filename: Optional filename to append
            use_subdir: Whether to create name subdirectory
            
        Returns:
            Full output path
            
        Examples:
            # Directory only with subdir
            >>> self.build_output_path('/out', name='cluster1')
            '/out/cluster1'
            
            # Directory + filename
            >>> self.build_output_path('/out', 'cluster1', 'vars.yaml')
            '/out/cluster1/vars.yaml'
            
            # Flat structure (no subdir)
            >>> self.build_output_path('/out', 'cluster1', 'vars.yaml', use_subdir=False)
            '/out/vars.yaml'
        """
        output_path = base_dir

        if name and use_subdir:
            output_path = os.path.join(output_path, name)

        if filename:
            output_path = os.path.join(output_path, filename)

        return output_path

    def get_runtime_dir(self, base_dir, cloud_provider, composition, use_cloud_subdir=True):
        """
        Get runtime directory path following standard structure.
        
        Builds consistent runtime paths for Terraform execution across different
        cloud providers and compositions.
        
        Args:
            base_dir: Base output directory
            cloud_provider: Cloud provider name (e.g., 'aws', 'gcp', 'azure')
            composition: Composition name (e.g., 'cluster', 'vpc', 'node-groups')
            use_cloud_subdir: Whether to include cloud provider in path structure
            
        Returns:
            Runtime directory path
            
        Examples:
            >>> self.get_runtime_dir('/tmp/runtime', 'aws', 'cluster')
            '/tmp/runtime/aws/cluster'
            
            >>> self.get_runtime_dir('/tmp/runtime', 'aws', 'vpc', use_cloud_subdir=False)
            '/tmp/runtime/vpc'
        """
        if use_cloud_subdir:
            return os.path.join(base_dir, cloud_provider, composition)
        else:
            return os.path.join(base_dir, composition)


    def generate_terraform_config(self, target_file, config_path, filtered_keys=None, excluded_keys=None,
                                  output_format="json", enclosing_key=None, print_data=False):
        """
        Generate Terraform configuration file from hiera.
        
        Generic method for generating any Terraform config file (provider, tfvars, etc.)
        
        Args:
            target_file: Path to output file
            config_path: Config path for hiera
            filtered_keys: Keys to include (None = all)
            excluded_keys: Keys to exclude (None = none)
            output_format: Output format (json or yaml)
            enclosing_key: Optional enclosing key (e.g., 'config')
            print_data: Whether to print generated data
            
        Returns:
            Path to generated file
        """
        filtered_keys = filtered_keys or []
        excluded_keys = excluded_keys or []

        # Don't log the command here - let the caller provide context

        self.generate_config(
            config_path=config_path,
            exclude_keys=excluded_keys,
            filters=filtered_keys,
            enclosing_key=enclosing_key,
            output_format=output_format,
            output_file=target_file,
            print_data=print_data,
            skip_interpolation_resolving=self.himl_args.skip_interpolation_resolving,
            skip_interpolation_validation=self.himl_args.skip_interpolation_validation,
            skip_secrets=self.himl_args.skip_secrets,
            multi_line_string=True
        )

        return target_file



"""
Validation runner for Kompos configurations.
Proactively checks for common configuration issues before generation.
"""

import logging
import os
import re
from typing import List, Dict, Any

from termcolor import colored

from kompos.parser import SubParserConfig
from kompos.runner import GenericRunner

logger = logging.getLogger(__name__)

RUNNER_TYPE = 'validate'


class ValidateParserConfig(SubParserConfig):
    def get_name(self):
        return RUNNER_TYPE

    def get_help(self):
        return 'Validate configuration for common issues before generation'

    def configure(self, parser):
        """Add validate-specific arguments. No HIML args needed."""
        parser.add_argument(
            '--rule',
            help='Run specific validation rule only',
            choices=['excluded-but-referenced', 'missing-layers', 'interpolation-syntax']
        )
        parser.add_argument(
            '--composition-type',
            help='Validate for specific composition type (e.g., account, cluster)',
        )
        parser.add_argument(
            '--strict',
            action='store_true',
            help='Exit with error code if any validation fails'
        )
        return parser


class ValidateRunner(GenericRunner):
    """
    Runner for validating Kompos configurations.
    
    Checks for common issues that would cause generation failures:
    - Keys referenced in interpolations but excluded from composition
    - Missing required configuration layers
    - Invalid composition type configurations
    - Circular dependencies
    
    Usage:
        kompos configs/path/to/composition validate [--rule RULE_NAME]
    """

    def __init__(self, kompos_config, config_path, execute=True):
        super().__init__(kompos_config, config_path, execute, RUNNER_TYPE)
        self.validation_rules = {
            'excluded-but-referenced': self._validate_excluded_but_referenced,
            'missing-layers': self._validate_missing_layers,
            'interpolation-syntax': self._validate_interpolation_syntax,
        }

    def run(self, args, extra_args=None):
        """Run validation checks"""
        logger.info(f"Running configuration validation for: {self.config_path}")

        # Determine which rules to run
        if args.rule:
            rules_to_run = {args.rule: self.validation_rules[args.rule]}
        else:
            rules_to_run = self.validation_rules

        # Run all validation rules
        all_results = []
        for rule_name, rule_func in rules_to_run.items():
            logger.info(f"Running validation rule: {rule_name}")
            try:
                results = rule_func(args)
                all_results.extend(results)
                logger.info(f"Rule {rule_name} completed with {len(results)} issue(s)")
            except Exception as e:
                logger.error(f"Rule {rule_name} failed: {e}", exc_info=True)
                raise

        # Display results
        self._display_results(all_results, args)

        # Return exit code
        if args.strict and any(r['severity'] == 'error' for r in all_results):
            return 1
        return 0

    def _validate_excluded_but_referenced(self, args) -> List[Dict[str, Any]]:
        """
        Validate that keys referenced in interpolations are not excluded.
        
        This catches the scenario where:
        1. A config file contains {{key.subkey}}
        2. The key exists in the hierarchy (has a value)
        3. BUT the key is excluded for this composition type
        
        Returns list of validation results with severity, message, and fix suggestions.
        """
        issues = []

        # Determine composition type
        comp_type = getattr(args, 'composition_type', None) if hasattr(args, 'composition_type') else None
        if not comp_type:
            comp_match = re.search(r'composition=(\w+)', self.config_path)
            if comp_match:
                comp_type = comp_match.group(1)

        if not comp_type:
            logger.warning(f"Could not determine composition type from path: {self.config_path}")
            return issues

        # Get excluded keys for this composition type
        excluded_keys = (self.kompos_config.config
                         .get('compositions', {})
                         .get('config_keys', {})
                         .get('excluded', {})
                         .get(comp_type, []))

        if not excluded_keys:
            logger.debug(f"No excluded keys for composition type: {comp_type} - skipping validation")
            return issues

        # Find all interpolations in config files
        interpolations = self._find_all_interpolations_in_hierarchy()

        logger.debug("Found %d interpolations in hierarchy", len(interpolations))

        # Check each interpolation's first-level key
        checked_keys = set()  # Avoid duplicate checks
        for interp_data in interpolations:
            interpolation = interp_data['interpolation']
            key_match = re.search(r'{{([^}]+)}}', interpolation)
            if not key_match:
                continue

            key_path = key_match.group(1).strip()
            first_key = key_path.split('.')[0]

            # Skip if already checked or not in excluded list
            if first_key in checked_keys or first_key not in excluded_keys:
                continue
            checked_keys.add(first_key)

            # Check if this key actually has a value in the hierarchy
            try:
                from kompos.runners.explore import ExploreRunner
                explore = ExploreRunner(
                    kompos_config=self.kompos_config,
                    config_path=self.config_path,
                    execute=False
                )
                trace = explore.trace_value(self.config_path, key_path, silent=True)

                # Check if key has a real value (not undefined)
                has_value = False
                last_value = None
                if trace and trace.get('trace'):
                    for step in trace['trace']:
                        val = step.get('value')
                        status = step.get('status', '').lower()
                        if val is not None and status not in ['undefined']:
                            has_value = True
                            last_value = val

                # If it has a value but is excluded, that's an issue
                if has_value:
                    # Find all files referencing this key
                    sources = [s for s in interpolations if s['interpolation'].startswith('{{' + first_key)]
                    source_files = list(set(s['file'] for s in sources))

                    issues.append({
                        'rule': 'excluded-but-referenced',
                        'severity': 'error',
                        'key': first_key,
                        'key_path': key_path,
                        'value': last_value,
                        'composition_type': comp_type,
                        'excluded_keys': excluded_keys,
                        'source_files': source_files[:5],  # Limit to 5 for display
                        'total_sources': len(source_files),
                        'message': (
                            f"Key '{first_key}' is referenced in config files but excluded "
                            f"for '{comp_type}' compositions"
                        ),
                        'fix_options': [
                            f"Remove '{first_key}' from .komposconfig.yaml exclusions for '{comp_type}'",
                            f"Move files using '{key_path}' to composition-specific defaults (e.g., defaults_{comp_type}.yaml)",
                            f"Remove unused interpolations containing '{first_key}' from global defaults"
                        ]
                    })
            except Exception as e:
                logger.debug(f"Could not trace key {key_path}: {e}")

        return issues

    def _validate_missing_layers(self, args) -> List[Dict[str, Any]]:
        """
        Validate that required hierarchy layers exist for composition.
        
        Placeholder for future implementation.
        """
        # TODO: Check if composition requires specific layers (region, etc)
        return []

    def _validate_interpolation_syntax(self, args) -> List[Dict[str, Any]]:
        """
        Validate interpolation syntax is correct.
        
        Placeholder for future implementation.
        """
        # TODO: Check for malformed interpolations, unclosed brackets, etc
        return []

    def _find_all_interpolations_in_hierarchy(self) -> List[Dict[str, str]]:
        """
        Find all interpolations in YAML files within the config hierarchy.
        
        Returns list of dicts with 'file', 'line', 'content', 'interpolation'
        """
        results = []

        # Start from config_path and walk up the hierarchy
        config_root = os.getcwd()  # Kompos runs from the project root

        # Convert config_path to absolute path
        if os.path.isabs(self.config_path):
            abs_config_path = self.config_path
        else:
            abs_config_path = os.path.join(config_root, self.config_path)

        current_dir = abs_config_path if os.path.isdir(abs_config_path) else os.path.dirname(abs_config_path)

        logger.debug(f"Searching for interpolations from: {current_dir}")
        logger.debug(f"Config root: {config_root}")

        visited_dirs = set()

        while current_dir and current_dir not in visited_dirs:
            visited_dirs.add(current_dir)

            # Stop if we've gone above the config root
            if not current_dir.startswith(config_root):
                logger.debug("Reached config root boundary, stopping")
                break

            logger.debug(f"Checking directory: {current_dir}")

            if os.path.isdir(current_dir):
                try:
                    yaml_files = [f for f in os.listdir(current_dir) if f.endswith(('.yaml', '.yml'))]
                    logger.debug(f"Found {len(yaml_files)} YAML files in {os.path.basename(current_dir)}")

                    for file in yaml_files:
                        file_path = os.path.join(current_dir, file)
                        try:
                            with open(file_path, 'r') as f:
                                for line_num, line in enumerate(f, 1):
                                    # Find all interpolations in this line
                                    for match in re.finditer(r'{{[^}]+}}', line):
                                        results.append({
                                            'file': os.path.relpath(file_path, config_root),
                                            'line': line_num,
                                            'content': line.strip(),
                                            'interpolation': match.group(0)
                                        })
                        except Exception as e:
                            logger.debug(f"Could not read file {file_path}: {e}")
                except Exception as e:
                    logger.warning(f"Could not list directory {current_dir}: {e}")

            # Move up one level
            parent = os.path.dirname(current_dir)
            if parent == current_dir:  # Reached root
                logger.debug("Reached filesystem root, stopping")
                break
            current_dir = parent

        logger.debug("Total interpolations found: %d", len(results))
        return results

    def _display_results(self, results: List[Dict[str, Any]], args):
        """Display validation results in a user-friendly format"""
        if not results:
            print(colored("✓ All validation checks passed!", 'green', attrs=['bold']))
            return

        # Count by severity
        errors = [r for r in results if r['severity'] == 'error']
        warnings = [r for r in results if r['severity'] == 'warning']

        print()
        print(colored("=" * 80, 'yellow'))
        print(colored("VALIDATION RESULTS", 'yellow', attrs=['bold']))
        print(colored("=" * 80, 'yellow'))
        print(f"Config Path: {colored(self.config_path, 'white', attrs=['bold'])}")
        print()

        if errors:
            print(colored(f"❌ {len(errors)} Error(s) Found", 'red', attrs=['bold']))
        if warnings:
            print(colored(f"⚠️  {len(warnings)} Warning(s) Found", 'yellow', attrs=['bold']))
        print()

        # Display each issue
        for i, issue in enumerate(results, 1):
            severity_color = 'red' if issue['severity'] == 'error' else 'yellow'
            severity_symbol = '❌' if issue['severity'] == 'error' else '⚠️'

            print(colored(f"{severity_symbol} Issue #{i}: {issue['rule']}", severity_color, attrs=['bold']))
            print(f"  {issue['message']}")
            print()

            # Display issue-specific details
            if issue['rule'] == 'excluded-but-referenced':
                print(f"  Key: {colored(issue['key'], 'cyan')}")
                print(f"  Current Value: {colored(issue['value'], 'white')}")
                print(f"  Composition Type: {colored(issue['composition_type'], 'cyan')}")
                print(f"  Excluded Keys: {colored(', '.join(issue['excluded_keys']), 'yellow')}")
                print()

                print(colored("  Referenced in:", 'white'))
                for src_file in issue['source_files']:
                    print(f"    • {colored(src_file, 'white', attrs=['dark'])}")
                if issue['total_sources'] > len(issue['source_files']):
                    remaining = issue['total_sources'] - len(issue['source_files'])
                    print(f"    ... and {remaining} more file(s)")
                print()

                print(colored("  Fix Options:", 'green', attrs=['bold']))
                for j, fix in enumerate(issue['fix_options'], 1):
                    print(f"    {j}. {colored(fix, 'green')}")
                print()

        print(colored("=" * 80, 'yellow'))
        print()



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os
import sys

from simpledi import Container, auto, cache, instance, ListInstanceProvider

from kompos.parser import RootParser
from kompos.helpers import print_error
from . import Executor
from .komposconfig import KomposConfig
from .runner import GenericRunner
from .runners.config import ConfigRenderParserConfig, ConfigRenderRunner
from .runners.explore import ExploreParserConfig, ExploreRunner
from .runners.helm import HelmParserConfig, HelmRunner
from .runners.helmfile import HelmfileParser, HelmfileRunner
from .runners.terraform import TerraformParser, TerraformRunner
from .runners.tfe import TFEParserConfig, TFERunner
from .runners.validate import ValidateParserConfig, ValidateRunner

logger = logging.getLogger(__name__)


class DebugBufferHandler(logging.Handler):
    """Capture all log messages for error context"""

    def __init__(self):
        super().__init__()
        self.buffer = []
        self.max_lines = 200  # Keep last 200 lines

    def emit(self, record):
        try:
            msg = self.format(record)
            self.buffer.append(msg)
            # Keep only last N lines
            if len(self.buffer) > self.max_lines:
                self.buffer.pop(0)
        except Exception:
            self.handleError(record)

    def get_buffer(self):
        return '\n'.join(self.buffer)

    def clear(self):
        self.buffer.clear()


# Global debug buffer
_debug_buffer = DebugBufferHandler()
_debug_buffer.setFormatter(logging.Formatter('%(levelname)s:%(name)s:%(message)s'))


def configure_logging(args):
    if args.verbose:
        if args.verbose > 1:
            # -vv: Show everything including himl debug
            logging.basicConfig(level=logging.DEBUG)
        else:
            # -v: Show INFO but suppress verbose himl logs
            logging.basicConfig(level=logging.INFO)
            logging.getLogger('himl.config_generator').setLevel(logging.WARNING)
    else:
        # No -v: Show INFO for kompos, suppress himl
        logging.basicConfig(level=logging.INFO)
        logging.getLogger('himl.config_generator').setLevel(logging.WARNING)

    # Always attach debug buffer to capture everything
    root_logger = logging.getLogger()
    root_logger.addHandler(_debug_buffer)


class AppContainer(Container):
    def __init__(self, argv=None):
        super(AppContainer, self).__init__()
        # Configure runners
        self.generic_runner = auto(GenericRunner)
        self.terraform_runner = auto(TerraformRunner)
        self.helm_runner = auto(HelmRunner)
        self.helmfile_runner = auto(HelmfileRunner)
        self.config_runner = auto(ConfigRenderRunner)
        self.explore_runner = auto(ExploreRunner)
        self.tfe_runner = auto(TFERunner)
        self.validate_runner = auto(ValidateRunner)

        # Configure parsers
        self.root_parser = auto(RootParser)
        parsers = ListInstanceProvider()
        parsers.add(auto(TerraformParser))
        parsers.add(auto(HelmParserConfig))
        parsers.add(auto(HelmfileParser))
        parsers.add(auto(ConfigRenderParserConfig))
        parsers.add(auto(ExploreParserConfig))
        parsers.add(auto(TFEParserConfig))
        parsers.add(auto(ValidateParserConfig))
        self.sub_parsers = parsers

        # Configure
        self.argv = instance(argv)
        self.configure()
        # bind the command executor
        self.execute = auto(Executor)

        # Set up kompos config
        self.kompos_config = cache(auto(KomposConfig))
        self.kompos_config.validate_version()
        self.kompos_config.vault_backend()

    def configure(self):
        args, extra_args = self.root_parser.parse_known_args(self.argv)

        configure_logging(args)
        logger.debug(f'cli args: {args}, extra_args: {extra_args}')

        # Bind some very useful dependencies
        self.package_dir = lambda c: os.path.dirname(__file__)
        self.console_args = cache(instance(args))
        self.console_extra_args = cache(instance(extra_args))
        self.command = lambda c: self.console_args.command
        self.root_path = cache(lambda c: get_root_path(c.console_args))
        self.config_path = cache(lambda c: get_config_path(c.console_args))

        # change path to the root_path
        logger.debug(f'root path: {self.root_path}')
        os.chdir(self.root_path)

        return args

    def run(self):
        # Check if command was provided
        if not self.console_args.command:
            print_error("No command specified")
            print("", file=sys.stderr)
            print("Available commands: tfe, terraform, helmfile, explore, validate, config", file=sys.stderr)
            print("", file=sys.stderr)
            print("Usage: kompos <config_path> <command> [options]", file=sys.stderr)
            print("", file=sys.stderr)
            print("Examples:", file=sys.stderr)
            print("  kompos configs/cloud=aws/.../composition=cluster tfe generate", file=sys.stderr)
            print("  kompos configs/cloud=aws/.../composition=account tfe generate", file=sys.stderr)
            print("  kompos configs/.../composition=cluster explore trace --key vpc.cidr", file=sys.stderr)
            print("", file=sys.stderr)
            print("Run 'kompos --help' for more information.", file=sys.stderr)
            sys.exit(1)
        
        command_name = f'{self.console_args.command}_runner'
        
        try:
            runner_instance = self.get_instance(command_name)
        except Exception as e:
            if "is not defined" in str(e):
                print_error(f"Unknown command: {self.console_args.command}")
                print("", file=sys.stderr)
                print("Available commands: tfe, terraform, helm, helmfile, explore, validate, config", file=sys.stderr)
                print("", file=sys.stderr)
                print("Run 'kompos --help' for more information.", file=sys.stderr)
                sys.exit(1)
            raise

        if not os.path.isdir(self.config_path):
            print_error(f"Invalid composition path: {self.config_path}", 
                       "Composition path must be a valid directory.")
            sys.exit(1)

        try:
            return runner_instance.run(self.console_args, self.console_extra_args)
        except Exception as e:
            error_msg = str(e)
            error_type = type(e).__name__
            
            # Handle interpolation errors
            is_interpolation_error = "Interpolation could not be resolved" in error_msg and "{{" in error_msg
            
            # Handle type errors from nested key access on strings
            is_type_error = (error_type == 'TypeError' and 
                           'string indices must be integers' in error_msg)
            
            # Handle YAML syntax errors
            is_yaml_error = error_type == 'ConstructorError' and 'unhashable key' in error_msg

            if is_interpolation_error:
                # Show error message first (in red)
                from termcolor import colored
                sys.stdout.write("\n")
                sys.stdout.write(colored("=" * 80, 'red') + "\n")
                sys.stdout.write(colored("ERROR: " + error_msg, 'red', attrs=['bold']) + "\n")
                sys.stdout.write(colored("=" * 80, 'red') + "\n")
                sys.stdout.write("\n")
                sys.stdout.flush()

                # Then run auto-debug analysis (this adds more logs to buffer)
                self._handle_interpolation_error(e)

                # Exit cleanly (no traceback)
                sys.exit(1)
            elif is_type_error:
                # Show helpful error for type mismatches in interpolation
                from termcolor import colored
                print_error("Interpolation type error: Trying to access nested key on a string value")
                print("", file=sys.stderr)
                print("This usually means:", file=sys.stderr)
                print("  • You're accessing 'foo.bar.baz' but 'foo.bar' is a string, not a dict", file=sys.stderr)
                print("  • Check your config hierarchy for the interpolation path", file=sys.stderr)
                print("", file=sys.stderr)
                print(f"Use 'kompos {self.config_path} explore trace --key <key>' to debug", file=sys.stderr)
                sys.exit(1)
            elif is_yaml_error:
                # Handle YAML syntax errors
                self._handle_yaml_error(e)
                sys.exit(1)
            else:
                # For other errors, raise normally (with traceback)
                raise

    def _handle_interpolation_error(self, error):
        """
        Centralized handler for interpolation errors.
        Automatically runs debug analysis when himl interpolation fails.
        """
        import re

        error_msg = str(error)
        if "Interpolation could not be resolved" not in error_msg or "{{" not in error_msg:
            return  # Not an interpolation error

        # Extract unresolved interpolation
        match = re.search(r'{{[^}]+}}', error_msg)
        unresolved = match.group(0) if match else None

        if not unresolved:
            return

        # Try to run automatic debug analysis
        try:
            # Create explore runner instance
            explore = ExploreRunner(
                kompos_config=self.kompos_config,
                config_path=self.config_path,
                execute=False
            )

            # Determine excluded keys based on command and composition type
            excluded_keys = self._get_excluded_keys_from_error(error_msg)

            # If no keys found from error (logs were suppressed), get from komposconfig
            if not excluded_keys:
                # Extract composition type from config path
                import re
                comp_match = re.search(r'composition=(\w+)', self.config_path)
                if comp_match:
                    comp_type = comp_match.group(1)
                    # Get excluded keys directly from komposconfig dict
                    excluded_keys = (self.kompos_config.config
                                     .get('compositions', {})
                                     .get('config_keys', {})
                                     .get('excluded', {})
                                     .get(comp_type, []))

            # Run interpolation analysis
            analysis = explore.analyze_interpolation(
                config_path=self.config_path,
                raw_config=None,  # Will search filesystem instead
                interpolation=unresolved,
                excluded_keys=excluded_keys
            )

            # Display simplified, focused error
            if analysis and analysis.get('unresolved'):
                output_text = explore._format_as_text(analysis)
                # Write directly to stdout to avoid any logger formatting
                sys.stdout.write("\n")
                sys.stdout.write("!" * 80 + "\n")
                sys.stdout.write("! AUTO-DEBUG: Analyzing interpolation error...\n")
                sys.stdout.write("!" * 80 + "\n")
                sys.stdout.write(output_text)
                sys.stdout.write("\n")
                sys.stdout.flush()

        except Exception as debug_error:
            logger.debug(f"Could not run debug analysis: {debug_error}", exc_info=True)

    def _handle_yaml_error(self, error):
        """
        Handler for YAML syntax errors with helpful guidance.
        """
        import re
        from termcolor import colored
        
        error_msg = str(error)
        
        # Extract file and line number
        file_match = re.search(r'in "([^"]+)", line (\d+)', error_msg)
        if file_match:
            yaml_file = file_match.group(1)
            line_num = int(file_match.group(2))
            
            print_error("YAML syntax error in config file")
            print("", file=sys.stderr)
            print(f"File: {yaml_file}", file=sys.stderr)
            print(f"Line: {line_num}", file=sys.stderr)
            
            if 'unhashable key' in error_msg:
                print("", file=sys.stderr)
                print(colored("╭─ Likely Cause " + "─" * 63 + "╮", 'yellow'), file=sys.stderr)
                print(colored("│ Unquoted interpolation starting with {{ or {              │", 'yellow'), file=sys.stderr)
                print(colored("│                                                             │", 'yellow'), file=sys.stderr)
                print(colored("│ ✗ BAD:  foo: {{bar.baz}}                                   │", 'yellow'), file=sys.stderr)
                print(colored("│ ✓ GOOD: foo: \"{{bar.baz}}\"                                 │", 'yellow'), file=sys.stderr)
                print(colored("╰─" + "─" * 77 + "╯", 'yellow'), file=sys.stderr)
            
            # Try to show context from file
            try:
                with open(yaml_file, 'r') as f:
                    lines = f.readlines()
                    start = max(0, line_num - 3)
                    end = min(len(lines), line_num + 2)
                    
                    print("", file=sys.stderr)
                    print("Content around the error:", file=sys.stderr)
                    for i in range(start, end):
                        line_prefix = colored(f"  {i+1:3d}: ", 'dim')
                        line_content = lines[i].rstrip()
                        
                        if i + 1 == line_num:
                            # Highlight the error line
                            print(colored(f"► {i+1:3d}: {line_content}", 'red', attrs=['bold']), file=sys.stderr)
                        else:
                            print(f"{line_prefix}{line_content}", file=sys.stderr)
            except Exception as read_error:
                logger.debug(f"Could not read file for context: {read_error}")
        else:
            print_error(f"YAML error: {error_msg}")

    def _get_excluded_keys_from_error(self, error_msg):
        """Extract excluded keys from himl output in error context"""
        excluded = []
        for line in error_msg.split('\n'):
            if 'Excluding key' in line:
                # Parse "INFO:himl.config_generator:Excluding key workspaces"
                parts = line.split('Excluding key')
                if len(parts) > 1:
                    key = parts[1].strip()
                    excluded.append(key)
        return excluded


def run(args=None):
    """ App entry point """
    app_container = AppContainer(args)
    sys.exit(app_container.run())


def get_config_path(console_args):
    config_path = console_args.config_path

    # If a file was provided (e.g., from tab completion), use its parent directory
    if os.path.isfile(config_path):
        config_path = os.path.dirname(config_path)
        logger.debug(f"Config path is a file, using parent directory: {config_path}")

    if not os.path.isdir(config_path):
        print_error(f"Config path does not exist: {config_path}")
        # Check if path looks like it might be valid but we're in wrong directory
        if '/' in config_path and not config_path.startswith('/'):
            print("This is a relative path. Are you in the correct directory?", file=sys.stderr)
            print(f"Current directory: {os.getcwd()}", file=sys.stderr)
        sys.exit(1)

    return config_path


def get_root_path(args):
    """ Either the root_path option or the current working dir """
    if args.root_path:
        root_path = os.path.realpath(args.root_path)
        if not os.path.isdir(root_path):
            print_error(f"Root directory does not exist: {root_path}")
            sys.exit(1)
        return root_path

    return os.path.realpath(os.getcwd())
